# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class SCPResult(Enum):
    """
    The SCP Result codes.
    """
    #: SCPCommand completed OK.
    RC_OK = 0x80
    #: Bad packet length.
    RC_LEN = 0x81
    #: Bad checksum.
    RC_SUM = 0x82
    #: Bad/invalid command.
    RC_CMD = 0x83
    #: Invalid arguments.
    RC_ARG = 0x84
    #: Bad port number.
    RC_PORT = 0x85
    #: Timeout.
    RC_TIMEOUT = 0x86
    #: No P2P route.
    RC_ROUTE = 0x87
    #: Bad CPU number.
    RC_CPU = 0x88
    #: destination dead.
    RC_DEAD = 0x89
    #: No free Shared Memory buffers.
    RC_BUF = 0x8a
    #: No reply to open.
    RC_P2P_NOREPLY = 0x8b
    #: Open rejected.
    RC_P2P_REJECT = 0x8c
    #: Destination busy.
    RC_P2P_BUSY = 0x8d
    #: Destination did not respond.
    RC_P2P_TIMEOUT = 0x8e
    #: Packet transmission failed.
    RC_PKT_TX = 0x8f



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Mapping, Optional, Tuple
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinnman.constants import SCP_SCAMP_PORT
from spinnman.transceiver.transceiver import Transceiver
from spinnman.connections.udp_packet_connections import UDPConnection
from .spalloc_state import SpallocState
from .spalloc_boot_connection import SpallocBootConnection
from .spalloc_eieio_connection import SpallocEIEIOConnection
from .spalloc_eieio_listener import SpallocEIEIOListener
from .spalloc_scp_connection import SpallocSCPConnection


class SpallocJob(object, metaclass=AbstractBase):
    """
    Represents a job in Spalloc.

    Don't make this yourself. Use :py:class:`SpallocClient` instead.
    """
    __slots__ = ()

    @abstractmethod
    def get_state(self, wait_for_change: bool = False) -> SpallocState:
        """
        Get the current state of the machine.

        :param bool wait_for_change: Whether to wait for a change in state

        :rtype: SpallocState
        """
        raise NotImplementedError()

    @abstractmethod
    def get_root_host(self) -> Optional[str]:
        """
        Get the IP address for talking to the machine.

        :return: The IP address, or ``None`` if not allocated.
        :rtype: str or None
        """
        raise NotImplementedError()

    @abstractmethod
    def get_connections(self) -> Dict[Tuple[int, int], str]:
        """
        Get the mapping from board coordinates to IP addresses.

        :return: (x,y)->IP mapping, or ``None`` if not allocated
        :rtype: dict(tuple(int,int), str) or None
        """
        raise NotImplementedError()

    @abstractmethod
    def connect_to_board(
            self, x: int, y: int,
            port: int = SCP_SCAMP_PORT) -> SpallocSCPConnection:
        """
        Open a connection to a particular board in the job.

        :param int x: X coordinate of the board's Ethernet-enabled chip
        :param int y: Y coordinate of the board's Ethernet-enabled chip
        :param int port: UDP port to talk to; defaults to the SCP port
        :return: A connection that talks to the board.
        :rtype: SpallocProxiedConnection
        """
        raise NotImplementedError()

    @abstractmethod
    def connect_for_booting(self) -> SpallocBootConnection:
        """
        Open a connection to a job's allocation so it can be booted.

        :return: a boot connection
        :rtype: SpallocBootConnection
        """
        raise NotImplementedError()

    @abstractmethod
    def open_eieio_connection(self, x: int, y: int) -> SpallocEIEIOConnection:
        """
        Open an EIEIO connection to a specific board in a job.

        :param int x:
            The X coordinate of the Ethernet-enabled chip to connect to
        :param int y:
            The Y coordinate of the Ethernet-enabled chip to connect to
        :return: an EIEIO connection with a board address bound
        :rtype: SpallocEIEIOConnection
        """
        raise NotImplementedError()

    @abstractmethod
    def open_eieio_listener_connection(self) -> SpallocEIEIOListener:
        """
        Open a listening EIEIO connection to the job's boards. Messages cannot
        be sent on this connection unless you say which board to send to, but
        they can be received from all boards. You can also get the *server*
        side connection information so you can program that into a tag.

        :return: an EIEIO connection with no board address bound
        :rtype: SpallocEIEIOListener
        """
        raise NotImplementedError()

    @abstractmethod
    def open_udp_listener_connection(self) -> UDPConnection:
        """
        Open a listening UDP connection to the job's boards. Messages cannot
        be sent on this connection unless you say which board to send to, but
        they can be received from all boards. You can also get the *server*
        side connection information so you can program that into a tag.

        :return: a UDP connection with no board address bound
        :rtype: UDPConnection
        """
        raise NotImplementedError()

    @abstractmethod
    def create_transceiver(self) -> Transceiver:
        """
        Create a transceiver that will talk to this job. The transceiver will
        only be configured to talk to the SCP ports of the boards of the job.

        :rtype: Transceiver
        """
        raise NotImplementedError()

    @abstractmethod
    def wait_for_state_change(self, old_state: SpallocState,
                              timeout: Optional[int] = None) -> SpallocState:
        """
        Wait until the allocation is not in the given old state.

        :param SpallocState old_state:
            The state that we are looking to change out of.
        :param timeout:
            The time to wait, or None to wait forever
        :type timeout: int or None
        :return: The state that the allocation is now in.

            .. note::
                If the machine gets destroyed, this will not wait for it.
        :rtype: SpallocState
        """
        raise NotImplementedError()

    @abstractmethod
    def wait_until_ready(self, timeout: Optional[int] = None,
                         n_retries: Optional[int] = None):
        """
        Wait until the allocation is in the ``READY`` state.

        :param timeout: The timeout or None to wait forever
        :type timeout: int or None
        :param n_retries:
            The number of times to retry, or None to retry forever
        :type n_retries: int or None
        :raises Exception: If the allocation is destroyed
        """
        raise NotImplementedError()

    @abstractmethod
    def destroy(self, reason: str = "finished"):
        """
        Destroy the job.

        :param str reason: Why the job is being destroyed.
        """
        raise NotImplementedError()

    @abstractmethod
    def where_is_machine(self, x: int, y: int) -> Optional[
            Tuple[int, int, int]]:
        """
        Get the *physical* coordinates of the board hosting the given chip.

        :param int x: Chip X coordinate
        :param int y: Chip Y coordinate
        :return: physical board coordinates (cabinet, frame, board), or
            ``None`` if there are no boards currently allocated to the job or
            the chip lies outside the allocation.
        :rtype: tuple(int,int,int) or None
        """
        raise NotImplementedError()

    @abstractmethod
    def get_session_credentials_for_db(self) -> Mapping[Tuple[str, str], str]:
        """
        Get the session credentials for the job to be written into a database

        .. note::
            May assume that there is a ``proxy_configuration`` table with
            ``kind``, ``name`` and ``value`` columns.
        """
        raise NotImplementedError()

    def __enter__(self):
        """
        Return self on entering context.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Handle exceptions by killing the job and logging the exception in the
        job's destroy reason.
        """
        try:
            self.destroy(str(exc_value))
        except Exception:  # pylint: disable=broad-except
            # Ignore this exception; there's not much we can do with it
            pass
        return None



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Optional

from spinnman.messages.spinnaker_boot import SpinnakerBootMessage
from spinnman.constants import UDP_BOOT_CONNECTION_DEFAULT_PORT

from .udp_connection import UDPConnection

_ANTI_FLOOD_DELAY = 0.1


class BootConnection(UDPConnection):
    """
    A connection to the SpiNNaker board that uses UDP to for booting.
    """
    __slots__ = ()
    _REPR_TEMPLATE = (
        "BootConnection(local_host={}, local_port={}, remote_host={}, "
        "remote_port={})")

    def __init__(self, remote_host: Optional[str] = None):
        """
        :param str remote_host:
            The remote host name or IP address to send packets to.  If not
            specified, the socket will be available for listening only, and
            will throw and exception if used for sending
        :raise SpinnmanIOException:
            If there is an error setting up the communication channel
        """
        super().__init__(remote_host=remote_host,
                         remote_port=UDP_BOOT_CONNECTION_DEFAULT_PORT)

    def send_boot_message(self, boot_message: SpinnakerBootMessage):
        """
        Sends a SpiNNaker boot message using this connection.

        :param SpinnakerBootMessage boot_message: The message to be sent
        :raise SpinnmanIOException:
            If there is an error sending the message
        """
        self.send(boot_message.bytestring)

        # Sleep between messages to avoid flooding the machine
        time.sleep(_ANTI_FLOOD_DELAY)

    def receive_boot_message(
            self, timeout: Optional[float] = None) -> SpinnakerBootMessage:
        """
        Receives a boot message from this connection.  Blocks until a
        message has been received, or a timeout occurs.

        :param float timeout:
            The time in seconds to wait for the message to arrive; if not
            specified, will wait forever, or until the connection is closed.
        :return: a boot message
        :rtype: SpinnakerBootMessage
        :raise SpinnmanIOException:
            If there is an error receiving the message
        :raise SpinnmanTimeoutException:
            If there is a timeout before a message is received
        :raise SpinnmanInvalidPacketException:
            If the received packet is not a valid SpiNNaker boot message
        :raise SpinnmanInvalidParameterException:
            If one of the fields of the SpiNNaker boot message is invalid
        """
        data = self.receive(timeout)
        return SpinnakerBootMessage.from_bytestring(data, 0)

    def __repr__(self):
        return self._REPR_TEMPLATE.format(
            self.local_ip_address, self.local_port,
            self.remote_ip_address, self.remote_port)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, List
from spinn_utilities.overrides import overrides
from spinnman.connections.udp_packet_connections import SCAMPConnection
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)


class RoundRobinConnectionSelector(ConnectionSelector):
    """
    A connection selector that just spreads work as evenly as possible.
    """
    __slots__ = (
        "_connections",
        "_next_connection_index")

    def __init__(self, connections: List[SCAMPConnection]):
        """
        :param list(SCAMPConnection) connections:
            The connections to be used
        """
        self._connections = connections
        self._next_connection_index = 0

    @overrides(ConnectionSelector.get_next_connection)
    def get_next_connection(self, message: Any) -> SCAMPConnection:
        index = self._next_connection_index
        self._next_connection_index = (index + 1) % len(self._connections)
        return self._connections[index]



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .adc_info import ADCInfo
from .bmp_connection_data import BMPConnectionData
from .chip_info import ChipInfo
from .chip_summary_info import ChipSummaryInfo
from .cpu_info import CPUInfo
from .cpu_infos import CPUInfos
from .diagnostic_filter import DiagnosticFilter
from .executable_targets import ExecutableTargets
from .heap_element import HeapElement
from .io_buffer import IOBuffer
from .machine_dimensions import MachineDimensions
from .p2p_table import P2PTable
from .router_diagnostics import RouterDiagnostics
from .version_info import VersionInfo

__all__ = ["ADCInfo", "BMPConnectionData", "ChipInfo", "ChipSummaryInfo",
           "CPUInfo", "CPUInfos", "DiagnosticFilter",
           "ExecutableTargets", "HeapElement", "IOBuffer", "MachineDimensions",
           "P2PTable", "RouterDiagnostics", "VersionInfo"]



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class DiagnosticFilterEmergencyRoutingStatus(Enum):
    """
    Emergency routing status flags for the diagnostic filters.

    .. note::
        Only one has to match for the counter to be incremented.
    """
    #: Packet is not emergency routed
    NORMAL = 0
    #: Packet is in first hop of emergency route; packet should also have been
    #: sent here by normal routing
    FIRST_STAGE_COMBINED = 1
    #: Packet is in first hop of emergency route; packet wouldn't have reached
    #: this router without emergency routing
    FIRST_STAGE = 2
    #: Packet is in last hop of emergency route and should now return to normal
    #: routing
    SECOND_STAGE = 3



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides

from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPHeader, SDPFlag

from .check_ok_response import CheckOKResponse


class FixedRouteInit(AbstractSCPRequest[CheckOKResponse]):
    """
    Sets a fixed route entry.
    """
    __slots__ = ()

    def __init__(self, x: int, y: int, entry: int, app_id: int = 0):
        """
        :param int x: The x-coordinate of the chip, between 0 and 255,
            this is not checked due to speed restrictions
        :param int y: The y-coordinate of the chip, between 0 and 255,
            this is not checked due to speed restrictions
        :param int entry: the fixed route entry converted for writing
        :param int app_id:
            The ID of the application with which to associate the routes.
            If not specified, defaults to 0.
        :raise SpinnmanInvalidParameterException:
            * If x is out of range
            * If y is out of range
        """
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_RTR),
            argument_1=(app_id << 8) | 3, argument_2=entry)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse("RouterInit", "CMD_RTR")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import List, Optional

from spinn_machine.tags import AbstractTag, ReverseIPTag, IPTag

from spinnman.messages.scp.impl.iptag_get import IPTagGet, IPTagGetResponse
from spinnman.messages.scp.impl.iptag_get_info import IPTagGetInfo
from spinnman.messages.scp.impl.iptag_get_info_response import (
    IPTagGetInfoResponse)
from spinnman.connections.udp_packet_connections import SCAMPConnection

from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)


class GetTagsProcess(AbstractMultiConnectionProcess):
    """
    Gets information about the tags over the provided connection.
    """
    __slots__ = (
        "_tags",
        "_tag_info")

    def __init__(self, connection_selector: ConnectionSelector):
        """
        :param ConnectionSelector connection_selector:
        """
        super().__init__(connection_selector)
        self._tag_info: Optional[IPTagGetInfoResponse] = None
        self._tags: List[Optional[AbstractTag]] = []

    def __handle_tag_info_response(self, response: IPTagGetInfoResponse):
        self._tag_info = response

    def __handle_get_tag_response(
            self, tag: int, board_address, response: IPTagGetResponse):
        if response.in_use:
            ip = response.ip_address
            host = f"{ip[0]}.{ip[1]}.{ip[2]}.{ip[3]}"
            if response.is_reverse:
                self._tags[tag] = ReverseIPTag(
                    board_address, tag,
                    response.rx_port, response.spin_chip_x,
                    response.spin_chip_y, response.spin_cpu,
                    response.spin_port)
            else:
                self._tags[tag] = IPTag(
                    board_address, response.sdp_header.source_chip_x,
                    response.sdp_header.source_chip_y, tag, host,
                    response.port, response.strip_sdp)

    def get_tags(self, connection: SCAMPConnection) -> List[AbstractTag]:
        """
        :param SCAMPConnection connection:
        :rtype:
            list(~spinn_machine.tags.IPTag or ~spinn_machine.tags.ReverseIPTag)
        """
        # Get the tag information, without which we cannot continue
        with self._collect_responses():
            self._send_request(IPTagGetInfo(
                connection.chip_x, connection.chip_y),
                self.__handle_tag_info_response)
        assert self._tag_info is not None

        # Get the tags themselves
        n_tags = self._tag_info.pool_size + self._tag_info.fixed_size
        self._tags = [None] * n_tags
        with self._collect_responses():
            for tag in range(n_tags):
                self._send_request(IPTagGet(
                    connection.chip_x, connection.chip_y, tag),
                    partial(
                        self.__handle_get_tag_response, tag,
                        connection.remote_ip_address))

        # Return the tags
        return [tag for tag in self._tags if tag is not None]



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import FrozenSet, Tuple
from spinn_utilities.abstract_base import AbstractBase, abstractmethod


class SpallocMachine(object, metaclass=AbstractBase):
    """
    Represents a Spalloc-controlled machine.

    Don't make this yourself. Use :py:class:`SpallocClient` instead.
    """
    __slots__ = ()

    @property
    @abstractmethod
    def name(self) -> str:
        """
        The name of the machine.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def tags(self) -> FrozenSet[str]:
        """
        The tags of the machine.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def width(self) -> int:
        """
        The width of the machine, in boards.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def height(self) -> int:
        """
        The height of the machine, in boards.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def dead_boards(self) -> list:
        """
        The dead or out-of-service boards of the machine.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def dead_links(self) -> list:
        """
        The dead or out-of-service links of the machine.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def area(self) -> Tuple[int, int]:
        """
        Area of machine, in boards.

        :return: width, height
        :rtype: tuple(int,int)
        """
        raise NotImplementedError

# TODO: Add in where-is operations
# Not yet done because not needed immediately for toolchain



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Generic, TypeVar
from spinn_utilities.abstract_base import (
    AbstractBase, abstractmethod)
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.connections.udp_packet_connections import (
    BMPConnection, SCAMPConnection)

#: Type of connections selected between.
#: :meta private:
Conn = TypeVar("Conn", SCAMPConnection, BMPConnection)


class ConnectionSelector(Generic[Conn], metaclass=AbstractBase):
    """
    A connection selector for multi-connection processes.
    """
    __slots__ = ()

    @abstractmethod
    def get_next_connection(
            self, message: AbstractSCPRequest) -> Conn:
        """
        Get the index of the  next connection for the process from a list
        of connections.

        :param AbstractSCPRequest message: The SCP message to be sent
        :rtype: SCAMPConnection
        """
        raise NotImplementedError



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import List, Optional
from spinn_machine.machine import Machine
from spinnman.model.enums import CPUState

_THREE_WORDS = struct.Struct("<3I")
_TWO_BYTES = struct.Struct("<BB")
_FOUR_BYTES = struct.Struct("<4B")
_EIGHTEEN_BYTES = struct.Struct("<18B")
_ONE_SHORT = struct.Struct("<H")


class ChipSummaryInfo(object):
    """
    Represents the chip summary information read via an SCP command.
    """
    __slots__ = [
        "_core_states",
        "_ethernet_ip_address",
        "_is_ethernet_available",
        "_largest_free_sdram_block",
        "_largest_free_sram_block",
        "_n_cores",
        "_n_free_multicast_routing_entries",
        "_nearest_ethernet_x",
        "_nearest_ethernet_y",
        "_parent_link",
        "_working_links",
        "_x", "_y"]

    def __init__(self, chip_summary_data: bytes, offset: int, x: int, y: int):
        """
        :param bytes chip_summary_data: The data from the SCP response
        :param int offset: The offset into the data where the data starts
        :param int x: The x-coordinate of the chip that this data is from
        :param int y: The y-coordinate of the chip that this data is from
        """
        (chip_summary_flags, self._largest_free_sdram_block,
            self._largest_free_sram_block) = _THREE_WORDS.unpack_from(
                chip_summary_data, offset)
        self._n_cores = chip_summary_flags & 0x1F
        self._working_links = [
            link for link in range(0, 6)
            if chip_summary_flags >> (8 + link) & 1 != 0]
        self._n_free_multicast_routing_entries = \
            (chip_summary_flags >> 14) & 0x7FF
        self._is_ethernet_available = bool(chip_summary_flags & (1 << 25))

        data_offset = offset + 12
        self._core_states = [
            CPUState(state) for state in
            _EIGHTEEN_BYTES.unpack_from(chip_summary_data, data_offset)]
        data_offset += 18

        self._x = x
        self._y = y

        self._nearest_ethernet_x, self._nearest_ethernet_y = 0, 0
        self._ethernet_ip_address: Optional[str] = None

        (self._nearest_ethernet_y, self._nearest_ethernet_x) = \
            _TWO_BYTES.unpack_from(chip_summary_data, data_offset)
        data_offset += 2

        ip = _FOUR_BYTES.unpack_from(chip_summary_data, data_offset)
        ethernet_ip_address = f"{ip[0]}.{ip[1]}.{ip[2]}.{ip[3]}"
        if ethernet_ip_address != "0.0.0.0":
            self._ethernet_ip_address = ethernet_ip_address
        data_offset += 4

        # In case the data hasn't been added in the version of SCAMP being used
        self._parent_link: Optional[int] = None
        if len(chip_summary_data) > data_offset:
            (self._parent_link, ) = _ONE_SHORT.unpack_from(
                chip_summary_data, data_offset)
            # The root chip will use the P2P "self", which is outside the range
            # of valid links, so check and skip this one
            if self._parent_link > len(Machine.LINK_ADD_TABLE):
                self._parent_link = None
            data_offset += 2

    @property
    def x(self) -> int:
        """
        The X-coordinate of the chip that this data is from.

        :rtype: int
        """
        return self._x

    @property
    def y(self) -> int:
        """
        The Y-coordinate of the chip that this data is from.

        :rtype: int
        """
        return self._y

    @property
    def n_cores(self) -> int:
        """
        The number of cores working on the chip (including monitors).

        :rtype: int
        """
        return self._n_cores

    @property
    def core_states(self) -> List[CPUState]:
        """
        The state of the cores on the chip (list of one per core).

        :rtype: list(CPUState)
        """
        return self._core_states

    @property
    def working_links(self) -> List[int]:
        """
        The IDs of the working links outgoing from this chip.

        :rtype: list(int)
        """
        return self._working_links

    @property
    def is_ethernet_available(self) -> bool:
        """
        Whether the Ethernet connection is available on this chip.

        :rtype: bool
        """
        return self._is_ethernet_available

    @property
    def n_free_multicast_routing_entries(self) -> int:
        """
        The number of multicast routing entries free on this chip.

        :rtype: int
        """
        return self._n_free_multicast_routing_entries

    @property
    def largest_free_sdram_block(self) -> int:
        """
        The size of the largest block of free SDRAM in bytes.

        :rtype: int
        """
        return self._largest_free_sdram_block

    @property
    def largest_free_sram_block(self) -> int:
        """
        The size of the largest block of free SRAM in bytes.

        :rtype: int
        """
        return self._largest_free_sram_block

    @property
    def nearest_ethernet_x(self) -> int:
        """
        The X-coordinate of the nearest Ethernet chip.

        :rtype: int
        """
        return self._nearest_ethernet_x

    @property
    def nearest_ethernet_y(self) -> int:
        """
        The Y-coordinate of the nearest Ethernet chip.

        :rtype: int
        """
        return self._nearest_ethernet_y

    @property
    def ethernet_ip_address(self) -> Optional[str]:
        """
        The IP address of the Ethernet if up, or `None` if not.

        :rtype: str
        """
        return self._ethernet_ip_address

    def clear_ethernet_ip_address(self) -> None:
        """
        Forces the Ethernet IP address to `None`, in case of an errant chip.
        """
        self._ethernet_ip_address = None

    @property
    def parent_link(self) -> Optional[int]:
        """
        The link to the parent of the chip in the tree of chips from root.

        :rtype: int
        """
        return self._parent_link

    def __repr__(self) -> str:
        return f"x:{self.x} y:{self.y} n_cores:{self.n_cores}"



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


# pylint: disable=invalid-name
class SDP_PORTS(Enum):
    """
    SDP port handling output buffering data streaming.
    """

    #: Command port for the buffered in functionality.
    INPUT_BUFFERING_SDP_PORT = 1
    #: Command port for the buffered out functionality.
    OUTPUT_BUFFERING_SDP_PORT = 2
    #: Command port for resetting runtime, etc.
    #: See :py:class:`SDP_RUNNING_MESSAGE_CODES`
    RUNNING_COMMAND_SDP_PORT = 3
    #: Extra monitor core reinjection control protocol.
    #: See :py:class:`ReinjectorSCPCommands`
    EXTRA_MONITOR_CORE_REINJECTION = 4
    #: Extra monitor core outbound data transfer protocol
    EXTRA_MONITOR_CORE_DATA_SPEED_UP = 5
    #: Extra monitor core inbound data transfer protocol
    #: See :py:class:`SpeedupInSCPCommands`
    EXTRA_MONITOR_CORE_DATA_IN_SPEED_UP = 6



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Generic, TypeVar
from spinn_utilities.overrides import overrides
from spinnman.connections.udp_packet_connections import (
    SCAMPConnection, BMPConnection)
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)

#: Type of connections selected between.
#: :meta private:
Conn = TypeVar("Conn", SCAMPConnection, BMPConnection)


class FixedConnectionSelector(ConnectionSelector[Conn], Generic[Conn]):
    """
    A connection selector that only uses a single connection.
    """
    __slots__ = ("__connection", )

    def __init__(self, connection: Conn):
        """
        :param SCAMPConnection connection:
            The connection to be used
        """
        self.__connection: Conn = connection

    @overrides(ConnectionSelector.get_next_connection)
    def get_next_connection(self, message: Any) -> Conn:
        return self.__connection



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Optional, overload
from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, AbstractSCPResponse)
from spinnman.messages.scp.enums import AllocFree, SCPCommand, SCPResult
from spinnman.messages.sdp import SDPFlag, SDPHeader
from spinnman.exceptions import SpinnmanUnexpectedResponseCodeException

_ONE_WORD = struct.Struct("<I")


class _SCPSDRAMDeAllocResponse(AbstractSCPResponse):
    """
    An SCP response to a request to deallocate SDRAM.
    """
    __slots__ = (
        "_number_of_blocks_freed",
        "_read_n_blocks_freed")

    def __init__(self, read_n_blocks_freed: bool = False):
        """
        """
        super().__init__()
        self._number_of_blocks_freed = 0
        self._read_n_blocks_freed = read_n_blocks_freed

    @overrides(AbstractSCPResponse.read_data_bytestring)
    def read_data_bytestring(self, data: bytes, offset: int):
        result = self.scp_response_header.result
        if result != SCPResult.RC_OK:
            raise SpinnmanUnexpectedResponseCodeException(
                "SDRAM deallocation", "CMD_DEALLOC", result.name)
        if self._read_n_blocks_freed:
            self._number_of_blocks_freed = _ONE_WORD.unpack_from(
                data, offset)[0]

            # check that the base address is not null (0 in python case) as
            # this reflects a issue in command on SpiNNaker side
            if self._number_of_blocks_freed == 0:
                raise SpinnmanUnexpectedResponseCodeException(
                    "SDRAM deallocation response base address", "CMD_DEALLOC",
                    result.name)

    @property
    def number_of_blocks_freed(self) -> int:
        """
        The number of allocated blocks that have been freed from the
        app_id given.

        :rtype: int
        """
        return self._number_of_blocks_freed


class SDRAMDeAlloc(AbstractSCPRequest[_SCPSDRAMDeAllocResponse]):
    """
    An SCP Request to free space in the SDRAM.
    """
    __slots__ = "_read_n_blocks_freed",

    @overload
    def __init__(self, x: int, y: int, *, app_id: int,
                 base_address: None = None):
        ...

    @overload
    def __init__(self, x: int, y: int, *, app_id: None = None,
                 base_address: int):
        ...

    def __init__(self, x: int, y: int, *, app_id: Optional[int] = None,
                 base_address: Optional[int] = None):
        """
        :param int x:
            The x-coordinate of the chip to allocate on, between 0 and 255
        :param int y:
            The y-coordinate of the chip to allocate on, between 0 and 255
        :param int app_id: The ID of the application, between 0 and 255
        :param base_address: The start address in SDRAM to which the block
            needs to be deallocated, or `None` if deallocating via app_id
        :type base_address: int or None
        """
        # pylint: disable=unsupported-binary-operation
        if base_address is not None:
            assert app_id is None
            super().__init__(
                SDPHeader(
                    flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                    destination_cpu=0, destination_chip_x=x,
                    destination_chip_y=y),
                SCPRequestHeader(command=SCPCommand.CMD_ALLOC),
                argument_1=(
                    AllocFree.
                    FREE_SDRAM_BY_POINTER.value),
                argument_2=base_address)
            self._read_n_blocks_freed = False
        else:
            assert app_id is not None
            super().__init__(
                SDPHeader(
                    flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                    destination_cpu=0, destination_chip_x=x,
                    destination_chip_y=y),
                SCPRequestHeader(command=SCPCommand.CMD_ALLOC),
                argument_1=(
                    app_id << 8 |
                    AllocFree.
                    FREE_SDRAM_BY_APP_ID.value))
            self._read_n_blocks_freed = True

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> _SCPSDRAMDeAllocResponse:
        return _SCPSDRAMDeAllocResponse(self._read_n_blocks_freed)



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from spinn_utilities.log import FormatAdapter
from spinn_utilities.overrides import overrides
from spinn_machine.data.machine_data_writer import MachineDataWriter
from spinnman.transceiver import Transceiver
from .spinnman_data_view import _SpiNNManDataModel, SpiNNManDataView

logger = FormatAdapter(logging.getLogger(__name__))
__temp_dir = None

REPORTS_DIRNAME = "reports"
# pylint: disable=protected-access


class SpiNNManDataWriter(MachineDataWriter, SpiNNManDataView):
    """
    See :py:class:`~spinn_utilities.data.UtilsDataWriter`

    This class is designed to only be used directly within the SpiNNMan
    repository unit tests as all methods are available to subclasses
    """
    __data = _SpiNNManDataModel()
    __slots__ = ()

    def _spinnman_mock(self) -> None:
        """
        Like :py:meth:`_mock` but does not call superclass `_mock`.
        This method should only be called by `mock` via `_mock`
        """
        self.__data._clear()

    @overrides(MachineDataWriter._mock)
    def _mock(self) -> None:
        MachineDataWriter._mock(self)
        self._spinnman_mock()

    def _spinnman_setup(self) -> None:
        """
        Like :py:meth:`_setup` but does not call superclass `_setup`.

        This method should only be called by `setup` (via `_setup`)
        """
        self.__data._clear()

    @overrides(MachineDataWriter._setup)
    def _setup(self) -> None:
        MachineDataWriter._setup(self)
        self._spinnman_setup()

    def _local_hard_reset(self) -> None:
        """
        Puts spinnman data back into the state expected at graph changed and
        `sim.reset`.

        Unlike hard_reset this method does not call super classes

        This method should only be called by hard_reset (via _hard_reset)
        """

        self.__data._hard_reset()

    @overrides(MachineDataWriter._hard_reset)
    def _hard_reset(self) -> None:
        MachineDataWriter._hard_reset(self)
        self._local_hard_reset()

    def _local_soft_reset(self) -> None:
        """
        Puts all data back into the state expected at `sim.reset` but not
        graph changed.

        Unlike soft_reset this method does not call superclasses

        This method should only be called by soft_reset (via _soft_reset)
        """
        self.__data._soft_reset()

    @overrides(MachineDataWriter._soft_reset)
    def _soft_reset(self) -> None:
        MachineDataWriter._soft_reset(self)
        self._local_soft_reset()

    def set_transceiver(self, transceiver: Transceiver):
        """
        Sets the transceiver object.

        :param Transceiver transceiver:
        :raises TypeError: If the transceiver is not a Transceiver subclass
        """
        if not isinstance(transceiver, Transceiver):
            raise TypeError("transceiver should be a Transceiver subclass")
        if self.__data._transceiver:
            raise NotImplementedError(
                "Over writing and existing transceiver not supported")
        self.__data._transceiver = transceiver



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand, Signal
from spinnman.messages.sdp import SDPFlag, SDPHeader
from .check_ok_response import CheckOKResponse

_APP_MASK = 0xFF


def _get_data(app_id, signal):
    data = (_APP_MASK << 8) | app_id
    data += signal.value << 16
    return data


class AppStop(AbstractSCPRequest[CheckOKResponse]):
    """
    An SCP Request to stop an application.
    """
    __slots__ = ()

    def __init__(self, app_id: int):
        """
        :param int app_id: The ID of the application, between 0 and 255
        """
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0,
                destination_chip_x=self.DEFAULT_DEST_X_COORD,
                destination_chip_y=self.DEFAULT_DEST_Y_COORD),
            SCPRequestHeader(command=SCPCommand.CMD_NNP),
            argument_1=(0x3f << 16),
            argument_2=(5 << 28) | _get_data(app_id, Signal.STOP),
            argument_3=(1 << 31) + (0x3f << 8))

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse("Send Stop", "CMD_NNP")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Collection

from spinn_machine import Router
from spinn_machine.multicast_routing_entry import MulticastRoutingEntry

from spinnman.exceptions import SpinnmanInvalidParameterException
from spinnman.messages.scp.impl import RouterInit, RouterAlloc
from spinnman.messages.scp.impl.router_alloc import RouterAllocResponse

from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)
from .write_memory_process import WriteMemoryProcess

_ROUTE_PATTERN = struct.Struct("<H2xIII")
_END_PATTERN = struct.Struct("<IIII")
_TABLE_ADDRESS = 0x67800000


class LoadMultiCastRoutesProcess(AbstractMultiConnectionProcess):
    """
    A process for loading the multicast routing table on a SpiNNaker chip.
    """
    __slots__ = ("_base_address", )

    def __init__(self, connection_selector: ConnectionSelector):
        """
        :param ConnectionSelector connection_selector:
        """
        super().__init__(connection_selector)
        self._base_address = 0

    def __handle_router_alloc_response(self, response: RouterAllocResponse):
        self._base_address = response.base_address

    def load_routes(
            self, x: int, y: int, routes: Collection[MulticastRoutingEntry],
            app_id: int):
        """
        :param int x:
        :param int y:
        :param list(~spinn_machine.MulticastRoutingEntry) routes:
        :param int app_id:
        """
        # Create the routing data - 16 bytes per entry plus one for the end
        # entry
        routing_data = bytearray(16 * (len(routes) + 1))
        n_entries = 0
        for route in routes:
            route_entry = \
                Router.convert_routing_table_entry_to_spinnaker_route(route)

            _ROUTE_PATTERN.pack_into(
                routing_data, n_entries * 16, n_entries,
                route_entry, route.key, route.mask)
            n_entries += 1

        # Add an entry to mark the end
        _END_PATTERN.pack_into(
            routing_data, n_entries * 16,
            0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF)

        # Upload the data
        process = WriteMemoryProcess(self._conn_selector)
        process.write_memory_from_bytearray(
            (x, y, 0), _TABLE_ADDRESS, routing_data, 0, len(routing_data))

        # Allocate space in the router table
        with self._collect_responses():
            self._send_request(RouterAlloc(x, y, app_id, n_entries),
                               self.__handle_router_alloc_response)
        if self._base_address == 0:
            raise SpinnmanInvalidParameterException(
                "Allocation base address", str(self._base_address),
                "Not enough space to allocate the entries")

        # Load the entries
        with self._collect_responses():
            self._send_request(RouterInit(
                x, y, n_entries, _TABLE_ADDRESS, self._base_address, app_id))



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-many-arguments
from contextlib import contextmanager
import io
import os
import logging
import struct
import time
from typing import Optional
from spinn_utilities.overrides import overrides
from spinn_utilities.abstract_base import (
    AbstractBase, abstractmethod)
from spinn_utilities.log import FormatAdapter
from spinn_utilities.logger_utils import warn_once
from spinn_utilities.require_subclass import require_subclass
from spinn_machine import CoreSubsets
from spinnman.constants import (
    ROUTER_REGISTER_BASE_ADDRESS, ROUTER_FILTER_CONTROLS_OFFSET,
    ROUTER_DIAGNOSTIC_FILTER_SIZE)
from spinnman.exceptions import SpinnmanException
from spinnman.extended import (
    BMPSetLed, DeAllocSDRAMProcess, ReadADC, SetLED, WriteMemoryFloodProcess)
from spinnman.model import DiagnosticFilter
from spinnman.model.enums import CPUState
from spinnman.messages.scp.enums import Signal
from spinnman.messages.scp.impl import (
    ReadMemory, ApplicationRun)
from spinnman.connections.udp_packet_connections import SCAMPConnection
from spinnman.constants import SYSTEM_VARIABLE_BASE_ADDRESS
from spinnman.data import SpiNNManDataView
from spinnman.messages.spinnaker_boot import SystemVariableDefinition
from spinnman.processes import (
    ConnectionSelector, GetHeapProcess, ReadMemoryProcess,
    SendSingleCommandProcess, WriteMemoryProcess)
from spinnman.transceiver.extendable_transceiver import ExtendableTransceiver

_ONE_BYTE = struct.Struct("B")
_ONE_WORD = struct.Struct("<I")

logger = FormatAdapter(logging.getLogger(__name__))


@require_subclass(ExtendableTransceiver)
class ExtendedTransceiver(object, metaclass=AbstractBase):
    """
    Allows a Transceiver to support extra method not currently needed.

    All methods here are in danger of being removed if they become too hard
    to support and many are untested so use at your own risk.
    It is undetermined if these will work with Spin2 boards.

    If any method here is considered important to keep please move it to
    Transceiver and its implementations
    """
    __slots__ = ()

    # calls many methods only reachable do to require_subclass
    # pylint: disable=no-member,assigning-non-slot
    # pylint: disable=access-member-before-definition
    # pylint: disable=attribute-defined-outside-init
    # pylint: disable=protected-access

    @abstractmethod
    def _where_is_xy(self, x: int, y: int) -> Optional[str]:
        raise NotImplementedError

    @property
    @abstractmethod
    @overrides(ExtendableTransceiver.scamp_connection_selector)
    def scamp_connection_selector(self) -> ConnectionSelector:
        """
        Returns the scamp selector

        :rtype: AbstractMultiConnectionProcessConnectionSelector
        """
        raise NotImplementedError

    def is_connected(self, connection=None):
        """
        Determines if the board can be contacted via SCAMP

        :param Connection connection:
            The connection which is to be tested.  If `None`,
            all Scamp connections will be tested,
            and the board will be considered
            to be connected if any one connection works.
        :return: True if the board can be contacted, False otherwise
        :rtype: bool
        """
        if connection is not None:
            return connection.is_connected()
        return any(c.is_connected() and isinstance(c, SCAMPConnection)
                   for c in self.scamp_connections)

    def get_iobuf_from_core(self, x, y, p):
        """
        Get the contents of IOBUF for a given core.

        .. warning::
            This method is currently deprecated and likely to be removed.

        :param int x: The x-coordinate of the chip containing the processor
        :param int y: The y-coordinate of the chip containing the processor
        :param int p: The ID of the processor to get the IOBUF for
        :return: An IOBUF buffer
        :rtype: IOBuffer
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If chip_and_cores contains invalid items
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        warn_once(logger, "The get_iobuf_from_core method is deprecated and "
                  "likely to be removed.")
        core_subsets = CoreSubsets()
        core_subsets.add_processor(x, y, p)
        return next(self.get_iobuf(core_subsets))

    @contextmanager
    def _chip_execute_lock(self, x, y):
        """
        Get a lock for executing an executable on a chip.

        .. warning::
            This method is currently deprecated and untested as there is no
            known use except for execute, which is itself deprecated.

        :param int x:
        :param int y:
        """
        # Check if there is a lock for the given chip
        with self._chip_execute_lock_condition:
            chip_lock = self._chip_execute_locks[x, y]
        # Acquire the lock for the chip
        chip_lock.acquire()

        # Increment the lock counter (used for the flood lock)
        with self._chip_execute_lock_condition:
            self._n_chip_execute_locks += 1

        try:
            yield chip_lock
        finally:
            with self._chip_execute_lock_condition:
                # Release the chip lock
                chip_lock.release()
                # Decrement the lock and notify
                self._n_chip_execute_locks -= 1
                self._chip_execute_lock_condition.notify_all()

    def execute(self, x, y, processors, executable, app_id, n_bytes=None,
                wait=False):
        """
        Start an executable running on a single chip.

        .. warning::
            This method is currently deprecated and likely to be removed.

        :param int x:
            The x-coordinate of the chip on which to run the executable
        :param int y:
            The y-coordinate of the chip on which to run the executable
        :param list(int) processors:
            The cores on the chip on which to run the application
        :param executable:
            The data that is to be executed. Should be one of the following:

            * An instance of RawIOBase
            * A bytearray/bytes
            * A filename of a file containing the executable (in which case
              `is_filename` must be set to True)
        :type executable:
            ~io.RawIOBase or bytes or bytearray or str
        :param int app_id:
            The ID of the application with which to associate the executable
        :param int n_bytes:
            The size of the executable data in bytes. If not specified:

            * If executable is an RawIOBase, an error is raised
            * If executable is a bytearray, the length of the bytearray will
              be used
            * If executable is an int, 4 will be used
            * If executable is a str, the length of the file will be used
        :param bool wait:
            True if the binary should enter a "wait" state on loading
        :raise SpinnmanIOException:
            * If there is an error communicating with the board
            * If there is an error reading the executable
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If x, y, p does not lead to a valid core
            * If app_id is an invalid application ID
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        warn_once(logger, "The Transceiver's execute method is deprecated "
                          "likely to be removed.")
        # Lock against updates
        with self._chip_execute_lock(x, y):
            # Write the executable
            self.write_memory(
                x, y, self._EXECUTABLE_ADDRESS, executable, n_bytes=n_bytes)

            # Request the start of the executable
            process = SendSingleCommandProcess(self._scamp_connection_selector)
            process.execute(ApplicationRun(app_id, x, y, processors, wait))

    def execute_application(self, executable_targets, app_id):
        """
        Execute a set of binaries that make up a complete application on
        specified cores, wait for them to be ready and then start all of the
        binaries.

        .. note::
            This will get the binaries into c_main but will not signal the
            barrier.

        :param ExecutableTargets executable_targets:
            The binaries to be executed and the cores to execute them on
        :param int app_id: The app_id to give this application
        """
        # Execute each of the binaries and get them in to a "wait" state
        for binary in executable_targets.binaries:
            core_subsets = executable_targets.get_cores_for_binary(binary)
            self.execute_flood(core_subsets, binary, app_id, wait=True)

        # Sleep to allow cores to get going
        time.sleep(0.5)

        # Check that the binaries have reached a wait state
        count = self.get_core_state_count(app_id, CPUState.READY)
        if count < executable_targets.total_processors:
            cores_ready = self.get_cpu_infos(
                executable_targets.all_core_subsets, [CPUState.READY],
                include=False)
            if len(cores_ready) > 0:
                raise SpinnmanException(
                    f"Only {count} of {executable_targets.total_processors} "
                    "cores reached ready state: "
                    f"{cores_ready.get_status_string()}")

        # Send a signal telling the application to start
        self.send_signal(app_id, Signal.START)

    def set_led(self, led, action, board):
        """
        Set the LED state of a board in the machine.

        .. warning::
            This method is currently deprecated and untested as there is no
            known use. Same functionality provided by ybug and bmpc.
            Retained in case needed for hardware debugging.

        :param led:
            Number of the LED or an iterable of LEDs to set the state of (0-7)
        :type led: int or iterable(int)
        :param LEDAction action:
            State to set the LED to, either on, off or toggle
        :param board: Specifies the board to control the LEDs of. This may
            also be an iterable of multiple boards. The
            command will actually be sent to the first board in the iterable.
        :type board: int or iterable(int)
        """
        warn_once(logger, "The set_led method is deprecated and "
                  "untested due to no known use.")
        process = SendSingleCommandProcess(self.bmp_selector)
        process.execute(BMPSetLed(led, action, board))

    def read_adc_data(self, board):
        """
        Read the BMP ADC data.

        .. warning::
            This method is currently deprecated and untested as there is no
            known use. Same functionality provided by ybug and bmpc.
            Retained in case needed for hardware debugging.

        :param int board: which board to request the ADC data from
        :return: the FPGA's ADC data object
        :rtype: ADCInfo
        """
        warn_once(logger, "The read_adc_data method is deprecated and "
                  "untested due to no known use.")
        process = SendSingleCommandProcess(self.bmp_selector)
        response = process.execute(ReadADC(board))
        return response.adc_info  # pylint: disable=no-member

    def write_neighbour_memory(self, x, y, link, base_address, data,
                               n_bytes=None, offset=0, cpu=0):
        """
        Write to the memory of a neighbouring chip using a LINK_READ SCP
        command. If sent to a BMP, this command can be used to communicate
        with the FPGAs' debug registers.

        .. warning::
            This method is deprecated and untested due to no known use.

        :param int x:
            The x-coordinate of the chip whose neighbour is to be written to
        :param int y:
            The y-coordinate of the chip whose neighbour is to be written to
        :param int link:
            The link index to send the request to (or if BMP, the FPGA number)
        :param int base_address:
            The address in SDRAM where the region of memory is to be written
        :param data: The data to write.  Should be one of the following:

            * An instance of RawIOBase
            * A bytearray/bytes
            * A single integer; will be written in little-endian byte order
        :type data:
            ~io.RawIOBase or bytes or bytearray or int
        :param int n_bytes:
            The amount of data to be written in bytes.  If not specified:

            * If `data` is an RawIOBase, an error is raised
            * If `data` is a bytearray, the length of the bytearray will be
              used
            * If `data` is an int, 4 will be used
        :param int offset:
            The offset where the valid data starts (if `data` is
            an int then offset will be ignored and used 0)
        :param int cpu:
            The CPU to use, typically 0 (or if a BMP, the slot number)
        :raise SpinnmanIOException:
            * If there is an error communicating with the board
            * If there is an error reading the data
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If `x, y` does not lead to a valid chip
            * If a packet is received that has invalid parameters
            * If `base_address` is not a positive integer
            * If `data` is an RawIOBase but `n_bytes` is not specified
            * If `data` is an int and `n_bytes` is more than 4
            * If `n_bytes` is less than 0
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        warn_once(logger, "The write_neighbour_memory method is deprecated "
                          "and untested due to no known use.")
        process = WriteMemoryProcess(self.scamp_connection_selector)
        if isinstance(data, io.RawIOBase):
            process.write_link_memory_from_reader(
                (x, y, cpu), link, base_address, data, n_bytes)
        elif isinstance(data, int):
            data_to_write = self._ONE_WORD.pack(data)
            process.write_link_memory_from_bytearray(
                (x, y, cpu), link, base_address, data_to_write, 0, 4)
        else:
            if n_bytes is None:
                n_bytes = len(data)
            process.write_link_memory_from_bytearray(
                (x, y, cpu), link, base_address, data, offset, n_bytes)

    def read_neighbour_memory(self, x, y, link, base_address, length, cpu=0):
        """
        Read some areas of memory on a neighbouring chip using a LINK_READ
        SCP command. If sent to a BMP, this command can be used to
        communicate with the FPGAs' debug registers.

        .. warning::
            This method is currently deprecated and untested as there is no
            known use. Same functionality provided by ybug and bmpc.
            Retained in case needed for hardware debugging.

        :param int x:
            The x-coordinate of the chip whose neighbour is to be read from
        :param int y:
            The y-coordinate of the chip whose neighbour is to be read from
        :param int cpu:
            The CPU to use, typically 0 (or if a BMP, the slot number)
        :param int link:
            The link index to send the request to (or if BMP, the FPGA number)
        :param int base_address:
            The address in SDRAM where the region of memory to be read starts
        :param int length: The length of the data to be read in bytes
        :return: An iterable of chunks of data read in order
        :rtype: bytes
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If one of `x`, `y`, `cpu`, `base_address` or `length` is invalid
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        try:
            warn_once(logger, "The read_neighbour_memory method is deprecated "
                      "and untested due to no known use.")
            process = ReadMemoryProcess(self.scamp_connection_selector)
            return process.read_link_memory(
                (x, y, cpu), link, base_address, length)
        except Exception:
            logger.info(self.where_is_xy(x, y))
            raise

    def _get_next_nearest_neighbour_id(self):
        with self._nearest_neighbour_lock:
            next_nearest_neighbour_id = (self._nearest_neighbour_id + 1) % 127
            self._nearest_neighbour_id = next_nearest_neighbour_id
        return next_nearest_neighbour_id

    def write_memory_flood(
            self, base_address, data, n_bytes=None, offset=0,
            is_filename=False):
        """
        Write to the SDRAM of all chips.

        :param int base_address:
            The address in SDRAM where the region of memory is to be written
        :param data:
            The data that is to be written.  Should be one of the following:

            * An instance of RawIOBase
            * A byte-string
            * A single integer
            * A file name of a file to read (in which case `is_filename`
              should be set to True)
        :type data:
            ~io.RawIOBase or bytes or bytearray or int or str
        :param int n_bytes:
            The amount of data to be written in bytes.  If not specified:

            * If `data` is an RawIOBase, an error is raised
            * If `data` is a bytearray or bytes, the length of the bytearray
              will be used
            * If `data` is an int, 4 will be used
            * If `data` is a str, the size of the file will be used
        :param int offset:
            The offset where the valid data starts; if `data` is
            an int, then the offset will be ignored and 0 is used.
        :param bool is_filename:
            True if `data` should be interpreted as a file name
        :raise SpinnmanIOException:
            * If there is an error communicating with the board
            * If there is an error reading the executable
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If one of the specified chips is not valid
            * If `app_id` is an invalid application ID
            * If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        process = WriteMemoryFloodProcess(self._scamp_connection_selector)
        # Ensure only one flood fill occurs at any one time
        with self._flood_write_lock:
            # Start the flood fill
            nearest_neighbour_id = self._get_next_nearest_neighbour_id()
            if isinstance(data, io.RawIOBase):
                process.write_memory_from_reader(
                    nearest_neighbour_id, base_address, data, n_bytes)
            elif isinstance(data, str) and is_filename:
                if n_bytes is None:
                    n_bytes = os.stat(data).st_size
                with open(data, "rb") as reader:
                    process.write_memory_from_reader(
                        nearest_neighbour_id, base_address, reader, n_bytes)
            elif isinstance(data, int):
                data_to_write = self._ONE_WORD.pack(data)
                process.write_memory_from_bytearray(
                    nearest_neighbour_id, base_address, data_to_write, 0)
            else:
                if n_bytes is None:
                    n_bytes = len(data)
                process.write_memory_from_bytearray(
                    nearest_neighbour_id, base_address, data, offset, n_bytes)

    def set_leds(self, x, y, cpu, led_states):
        """
        Set LED states.

        .. warning::
            The set_leds is deprecated and untested due to no known use.

        :param int x: The x-coordinate of the chip on which to set the LEDs
        :param int y: The x-coordinate of the chip on which to set the LEDs
        :param int cpu: The CPU of the chip on which to set the LEDs
        :param dict(int,int) led_states:
            A dictionary mapping SetLED index to state with
            0 being off, 1 on and 2 inverted.
        :raise SpinnmanIOException:
            If there is an error communicating with the board
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            If a packet is received that has invalid parameters
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        try:
            warn_once(logger, "The set_leds is deprecated and "
                      "untested due to no known use.")
            process = SendSingleCommandProcess(self._scamp_connection_selector)
            process.execute(SetLED(x, y, cpu, led_states))
        except Exception:
            logger.info(self.where_is_xy(x, y))
            raise

    def free_sdram(self, x, y, base_address):
        """
        Free allocated SDRAM.

        .. warning::
            This method is currently deprecated and likely to be removed.

        :param int x: The x-coordinate of the chip onto which to ask for memory
        :param int y: The y-coordinate of the chip onto which to ask for memory
        :param int base_address: The base address of the allocated memory
        """
        try:
            warn_once(logger, "The free_sdram method is deprecated and "
                      "likely to be removed.")
            process = DeAllocSDRAMProcess(self._scamp_connection_selector)
            process.de_alloc_sdram(x, y, base_address)
        except Exception:
            logger.info(self.where_is_xy(x, y))
            raise

    def free_sdram_by_app_id(self, x, y, app_id):
        """
        Free all SDRAM allocated to a given app ID.

        .. warning::
            This method is currently deprecated and untested as there is no
            known use. Same functionality provided by ybug and bmpc.
            Retained in case needed for hardware debugging.

        :param int x: The x-coordinate of the chip onto which to ask for memory
        :param int y: The y-coordinate of the chip onto which to ask for memory
        :param int app_id: The app ID of the allocated memory
        :return: The number of blocks freed
        :rtype: int
        """
        try:
            warn_once(logger, "The free_sdram_by_app_id method is deprecated "
                              "and untested due to no known use.")
            process = DeAllocSDRAMProcess(self._scamp_connection_selector)
            process.de_alloc_sdram(x, y, app_id)
            return process.no_blocks_freed
        except Exception:
            logger.info(self.where_is_xy(x, y))
            raise

    def get_router_diagnostic_filter(
            self, x: int, y: int, position: int) -> DiagnosticFilter:
        """
        Gets a router diagnostic filter from a router.

        :param int x:
            the X address of the router from which this filter is being
            retrieved
        :param int y:
            the Y address of the router from which this filter is being
            retrieved
        :param int position:
            the position in the list of filters to read the information from
        :return: The diagnostic filter read
        :rtype: DiagnosticFilter
        :raise SpinnmanIOException:
            * If there is an error communicating with the board
            * If there is an error reading the data
        :raise SpinnmanInvalidPacketException:
            If a packet is received that is not in the valid format
        :raise SpinnmanInvalidParameterException:
            * If x, y does not lead to a valid chip
            * If a packet is received that has invalid parameters
            * If position is less than 0 or more than 15
        :raise SpinnmanUnexpectedResponseCodeException:
            If a response indicates an error during the exchange
        """
        try:
            memory_position = (
                ROUTER_REGISTER_BASE_ADDRESS + ROUTER_FILTER_CONTROLS_OFFSET +
                position * ROUTER_DIAGNOSTIC_FILTER_SIZE)

            process: SendSingleCommandProcess = SendSingleCommandProcess(
                self.scamp_connection_selector)
            response = process.execute(
                ReadMemory((x, y, 0), memory_position, 4))
            return DiagnosticFilter.read_from_int(_ONE_WORD.unpack_from(
                response.data, response.offset)[0])
            # pylint: disable=no-member
        except Exception:
            logger.info(self._where_is_xy(x, y))
            raise

    @property
    def number_of_boards_located(self):
        """
        The number of boards currently configured.

        .. warning::
            This property is currently deprecated and likely to be removed.

        :rtype: int
        """
        warn_once(logger, "The number_of_boards_located method is deprecated "
                          "and likely to be removed.")
        if self.bmp_connection is not None:
            return max(1, len(self.bmp_connection.boards))
        else:
            # if no BMPs are available, then there's still at least one board
            return 1

    def get_heap(self, x, y, heap=SystemVariableDefinition.sdram_heap_address):
        """
        Get the contents of the given heap on a given chip.

        :param int x: The x-coordinate of the chip
        :param int y: The y-coordinate of the chip
        :param SystemVariableDefinition heap:
            The SystemVariableDefinition which is the heap to read
        :rtype: list(HeapElement)
        """
        try:
            process = GetHeapProcess(self.scamp_connection_selector)
            return process.get_heap((x, y), heap)
        except Exception:
            logger.info(self.where_is_xy(x, y))
            raise

    def __set_watch_dog_on_chip(self, x, y, watch_dog):
        """
        Enable, disable or set the value of the watch dog timer on a
        specific chip.

        .. warning::
            This method is currently deprecated and untested as there is no
            known use. Same functionality provided by ybug and bmpc.
            Retained in case needed for hardware debugging.

        :param int x: chip X coordinate to write new watchdog parameter to
        :param int y: chip Y coordinate to write new watchdog parameter to
        :param watch_dog:
            Either a Boolean indicating whether to enable (True) or
            disable (False) the watchdog timer, or an int value to set the
            timer count to
        :type watch_dog: bool or int
        """
        # build what we expect it to be
        warn_once(logger, "The set_watch_dog_on_chip method is deprecated "
                          "and untested due to no known use.")
        value_to_set = watch_dog
        watchdog = SystemVariableDefinition.software_watchdog_count
        if isinstance(watch_dog, bool):
            value_to_set = watchdog.default if watch_dog else 0

        # build data holder
        data = _ONE_BYTE.pack(value_to_set)

        # write data
        address = SYSTEM_VARIABLE_BASE_ADDRESS + watchdog.offset
        self.write_memory(x=x, y=y, base_address=address, data=data)

    def set_watch_dog(self, watch_dog):
        """
        Enable, disable or set the value of the watch dog timer.

        .. warning::
            This method is currently deprecated and untested as there is no
            known use. Same functionality provided by ybug and bmpc.
            Retained in case needed for hardware debugging.

        :param watch_dog:
            Either a Boolean indicating whether to enable (True) or
            disable (False) the watch dog timer, or an int value to set the
            timer count to.
        :type watch_dog: bool or int
        """
        warn_once(logger, "The set_watch_dog method is deprecated and "
                          "untested due to no known use.")
        for x, y in SpiNNManDataView.get_machine().chip_coordinates:
            self.__set_watch_dog_on_chip(x, y, watch_dog)



# Copyright (c) 2021 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Miscellaneous utilities for working with URLs relating to the Spalloc Client.
"""

from typing import Iterable, Tuple, Optional
from urllib.parse import urlparse, urlsplit, urlunparse


def clean_url(url: str) -> str:
    """
    Add a ``/`` to the end of the path part of a URL if there isn't one.

    :param str url:
    :rtype: str
    """
    r = urlparse(url)
    parts = list(r)
    # Add a / to the end of the path if it isn't there
    if not parts[2].endswith("/"):
        parts[2] += "/"
    return urlunparse(parts)


def parse_service_url(url: str) -> Tuple[str, Optional[str], Optional[str]]:
    """
    Parses a combined service reference. Must include a hostname.

    :param str url:
    :return: URL, username (may be `None`), password (may be `None`)
    :rtype: tuple(str,str,str)
    """
    pieces = urlparse(url)
    user = pieces.username
    password = pieces.password
    netloc = pieces.hostname
    if netloc is None:
        raise ValueError("URL must have a hostname")
    if pieces.port is not None:
        netloc += f":{pieces.port}"
    url = urlunparse((
        pieces.scheme, netloc, pieces.path, None, None, None))
    return url, user, password


def get_hostname(url: str) -> str:
    """
    Parses a URL and extracts the hostname part.
    A hostname must be present.
    """
    netloc = urlsplit(url).hostname
    if netloc is None:
        raise ValueError("URL must have a hostname")
    return netloc


def is_server_address(
        address: str, additional_schemes: Iterable[str] = ()) -> bool:
    """
    Test if the given address is a likely Spalloc server URL.

    :param str address: The address to check
    :param ~collections.abc.Iterable(str) additional_schemes:
        Any additional URL schemes that should be considered to be successes;
        typically ``{"spalloc"}`` when looser matching is required.
    :rtype: bool
    """
    schemes = {"http", "https"}
    if additional_schemes:
        schemes.update(additional_schemes)
    try:
        pieces = urlparse(address)
        scheme = pieces.scheme.lower()
        return scheme in schemes and pieces.netloc is not None
    except Exception:  # pylint: disable=broad-except
        return False



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import sys
import signal
import socket
from typing import Callable
from spinnman.connections.udp_packet_connections import IPAddressesConnection


def locate_connected_machine(handler: Callable[[str, float], bool]):
    """
    Locates any SpiNNaker machines IP addresses from the auto-transmitted
    packets from non-booted SpiNNaker machines.

    :param ~collections.abc.Callable handler:
        A callback that decides whether to stop searching. The callback is
        given two arguments: the IP address found and the current time. It
        should return True if the search should cease.
    """

    connection = IPAddressesConnection()
    seen_boards = set()
    while True:
        ip_address = connection.receive_ip_address()
        now = time.time()
        if ip_address is not None and ip_address not in seen_boards:
            seen_boards.add(ip_address)
            if handler(ip_address, now):
                break


if __name__ == "__main__":
    def _ctrlc_handler(sig, frame):
        """
        :return: Never returns as it causes a sys.exit()
        """
        # pylint: disable=unused-argument
        print("Exiting")
        sys.exit()

    def _print_connected(ip_address: str, timestamp: float) -> bool:
        try:
            hostname = f" ({socket.gethostbyaddr(ip_address)[0]})"
        except Exception:  # pylint: disable=broad-except
            hostname = ""
        print(f"{ip_address}{hostname} at {timestamp}")
        return False

    print("The following addresses might be SpiNNaker boards "
          "(press Ctrl-C to quit):")
    signal.signal(signal.SIGINT, _ctrlc_handler)
    locate_connected_machine(handler=_print_connected)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Kept for spalloc_server to use
def update_sdp_header_for_udp_send(sdp_header, source_x, source_y):
    """
    Apply defaults to the SDP header for sending over UDP.

    .. deprecated:: 7.0
        Use :py:meth:`SDPHeader.update_for_send` instead.

    :param SDPHeader sdp_header: The SDP header to update
    :param int source_x: Where the packet is deemed to originate: X coordinate
    :param int source_y: Where the packet is deemed to originate: Y coordinate
    """
    sdp_header.update_for_send(source_x, source_y)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct

from spinn_utilities.overrides import overrides
from spinnman.constants import EIEIO_COMMAND_IDS
from spinnman.messages.eieio import AbstractEIEIOMessage
from spinnman.messages.eieio.create_eieio_data import read_eieio_data_message

from .eieio_command_message import EIEIOCommandMessage
from .eieio_command_header import EIEIOCommandHeader


_PATTERN_BB = struct.Struct("<BB")


class HostSendSequencedData(EIEIOCommandMessage):
    """
    Packet sent from the host to the SpiNNaker system in the context of
    buffering input mechanism to identify packet which needs to be stored
    in memory for future use.
    """
    __slots__ = (
        "_eieio_data_message",
        "_region_id",
        "_sequence_no")

    def __init__(self, region_id: int, sequence_no: int,
                 eieio_data_message: AbstractEIEIOMessage):
        """

        :param int region_id:
        :param int sequence_no:
        :param AbstractEIEIOMessage eieio_data_message:
        """
        super().__init__(EIEIOCommandHeader(
            EIEIO_COMMAND_IDS.HOST_SEND_SEQUENCED_DATA))
        self._region_id = region_id
        self._sequence_no = sequence_no
        self._eieio_data_message = eieio_data_message

    @property
    def region_id(self) -> int:
        """
        The region_id passed into the init.

        :rtype: int
        """
        return self._region_id

    @property
    def sequence_no(self) -> int:
        """
        The sequence_no passed into the init.

        :rtype: int
        """
        return self._sequence_no

    @property
    def eieio_data_message(self) -> AbstractEIEIOMessage:
        """
        The eieio_data_message passed into the init.

        :return: AbstractEIEIOMessage
        """
        return self._eieio_data_message

    @staticmethod
    @overrides(EIEIOCommandMessage.get_min_packet_length)
    def get_min_packet_length() -> int:
        return 4

    @staticmethod
    @overrides(EIEIOCommandMessage.from_bytestring)
    def from_bytestring(command_header: EIEIOCommandHeader, data: bytes,
                        offset: int) -> "HostSendSequencedData":
        region_id, sequence_no = _PATTERN_BB.unpack_from(data, offset)
        eieio_data_message = read_eieio_data_message(data, offset)
        return HostSendSequencedData(
            region_id, sequence_no, eieio_data_message)

    @property
    @overrides(EIEIOCommandMessage.bytestring)
    def bytestring(self) -> bytes:
        return (super().bytestring +
                _PATTERN_BB.pack(self._region_id, self._sequence_no) +
                self._eieio_data_message.bytestring)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinn_utilities.typing.coords import XYP
from spinnman.constants import address_length_dtype
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from .check_ok_response import CheckOKResponse


class WriteMemory(AbstractSCPRequest[CheckOKResponse]):
    """
    A request to write memory on a chip.
    """
    __slots__ = "_data_to_write",

    def __init__(self, coordinates: XYP, base_address: int, data: bytes):
        """
        :param tuple(int,int,int) coordinates:
            The coordinates of the chip, X and Y between 0 and 255, and P
            between 0 and 17 (normally 0 when writing to SDRAM; may be other
            values when writing to ITCM or DTCM);
            these are not checked due to speed restrictions
        :param int base_address: The base_address to start writing to
            the base address is not checked to see if its not valid
        :param data: between 1 and 256 bytes of data to write;
            this is not checked due to speed restrictions
        :type data: bytearray or bytes
        """
        size = len(data)
        x, y, cpu = coordinates
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=cpu, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_WRITE),
            argument_1=base_address, argument_2=size,
            argument_3=address_length_dtype[
                (base_address % 4), (size % 4)].value,
            data=None)
        self._data_to_write = data

    @property
    def bytestring(self) -> bytes:
        return super().bytestring + bytes(self._data_to_write)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse("write to memory", "CMD_WRITE")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Generic, Optional, TypeVar, Set

from spinnman.constants import SCP_TIMEOUT

from spinnman.messages.scp.abstract_messages import AbstractSCPResponse
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums.scp_result import SCPResult

from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)
#: Type of responses.
#: :meta private:
R = TypeVar("R", bound=AbstractSCPResponse)


class SendSingleCommandProcess(AbstractMultiConnectionProcess, Generic[R]):
    """
    A process that sends a single command and waits for a simple response.
    """
    __slots__ = ("_response", )

    def __init__(self, connection_selector: ConnectionSelector,
                 n_retries: int = 3, timeout: float = SCP_TIMEOUT,
                 non_fail_retry_codes: Optional[Set[SCPResult]] = None):
        """
        :param ConnectionSelector connection_selector:
        :param int n_retries:
            The number of retries of a message to use. Passed to
            :py:class:`SCPRequestPipeLine`
        :param float timeout:
            The timeout, in seconds. Passed to
            :py:class:`SCPRequestPipeLine`
        :param Optional[Set[SCPResult]] non_fail_retry_codes:
            Optional set of responses that result in retry but after retrying
            don't then result in failure even if returned on the last call.
        """
        super().__init__(
            connection_selector, n_retries=n_retries, timeout=timeout,
            non_fail_retry_codes=non_fail_retry_codes)
        self._response: Optional[R] = None

    def __handle_response(self, response: R):
        self._response = response

    def execute(self, request: AbstractSCPRequest[R]) -> R:
        """
        :param AbstractSCPRequest request:
        :rtype: AbstractSCPResponse
        """
        with self._collect_responses():
            self._send_request(request, self.__handle_response)
        assert self._response is not None
        return self._response



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from .check_ok_response import CheckOKResponse

_IPTAG_SET = 1


class IPTagSet(AbstractSCPRequest[CheckOKResponse]):
    """
    An SCP Request to set an IP Tag.
    """
    __slots__ = ()

    def __init__(self, x: int, y: int, host, port: int, tag: int, *,
                 strip: bool = False, use_sender: bool = False):
        """
        :param int x: The x-coordinate of a chip, between 0 and 255
        :param int y: The y-coordinate of a chip, between 0 and 255
        :param host: The host address, as an array of 4 bytes
        :type host: bytearray or list(int)
        :param int port: The port, between 0 and 65535
        :param int tag: The tag, between 0 and 7
        :param bool strip: if the SDP header should be striped from the packet
        :param bool use_sender:
            if the sender IP address and port should be used
        """
        # pylint: disable=too-many-arguments
        strip_value = int(bool(strip))
        sender_value = int(bool(use_sender))
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=0, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_IPTAG),
            argument_1=(strip_value << 28) | (sender_value << 30) |
                       (_IPTAG_SET << 16) | tag,
            argument_2=port,
            argument_3=((host[3] << 24) | (host[2] << 16) |
                        (host[1] << 8) | host[0]))

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse("Set IP Tag", "CMD_IPTAG")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinnman.messages.eieio.data_messages import (
    EIEIODataMessage, EIEIODataHeader)


def read_eieio_data_message(data, offset):
    """
    Reads the content of an EIEIO data message and returns an object
    identifying the data which was contained in the packet.

    :param bytes data: data received from the network as a byte-string
    :param int offset: offset at which the parsing operation should start
    :return: an object which inherits from EIEIODataMessage which contains
        parsed data received from the network
    :rtype: EIEIODataMessage
    """
    eieio_header = EIEIODataHeader.from_bytestring(data, offset)
    offset += eieio_header.size
    return EIEIODataMessage(eieio_header, data, offset)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from spinn_utilities.overrides import overrides
from spinnman.messages.scp.abstract_messages import (
    AbstractSCPRequest, BMPRequest, BMPOKResponse)
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.enums import SCPCommand

_ONE_WORD = struct.Struct("<I")


# pylint: disable=wrong-spelling-in-docstring
class WriteFPGARegister(BMPRequest[BMPOKResponse]):
    """
    A request for writing a word to a FPGA (SPI) register.

    See the SpI/O project's spinnaker_fpga design's `README`_ for a listing
    of FPGA registers. The SpI/O project can be found on GitHub at:
    https://github.com/SpiNNakerManchester/spio/

    .. _README: https://github.com/SpiNNakerManchester/spio/\
                blob/master/designs/spinnaker_fpgas/README.md#spi-interface
    """
    __slots__ = ()

    def __init__(self, fpga_num: int, address: int, value: int, board: int):
        """
        :param int fpga_num: FPGA number (0, 1 or 2) to communicate with.
        :param int address: Register address to read or write to
            (will be rounded down to the nearest 32-bit word boundary).
        :param int value: A 32-bit int value to write to the register
        """
        super().__init__(
            board,
            SCPRequestHeader(command=SCPCommand.CMD_LINK_WRITE),
            argument_1=address & (~0x3), argument_2=4, argument_3=fpga_num,
            data=_ONE_WORD.pack(value))

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> BMPOKResponse:
        return BMPOKResponse(
            "Send FPGA register write", SCPCommand.CMD_LINK_WRITE)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.overrides import overrides
from spinn_utilities.typing.coords import XYP
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from .check_ok_response import CheckOKResponse


class WriteLink(AbstractSCPRequest[CheckOKResponse]):
    """
    A request to write memory on a neighbouring chip.
    """
    __slots__ = "_data_to_write",

    def __init__(
            self, coordinates: XYP, link: int, base_address: int, data: bytes):
        """
        :param tuple(int,int,int) coordinates:
            The coordinates of the core of the chip whose neighbour will be
            written to; X and Y between 0 and 255,
            CPU core normally 0 (or if a BMP then the board slot number)
        :param int link: The link number to write to between 0 and 5
            (or if a BMP, the FPGA between 0 and 2)
        :param int base_address: The base_address to start writing to
        :param bytes data: Up to 256 bytes of data to write
        """
        # pylint: disable=too-many-arguments
        x, y, cpu = coordinates
        super().__init__(
            SDPHeader(
                flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                destination_cpu=cpu, destination_chip_x=x,
                destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_LINK_WRITE),
            argument_1=base_address, argument_2=len(data), argument_3=link,
            data=None)
        self._data_to_write = data

    @property
    def bytestring(self) -> bytes:
        return super().bytestring + bytes(self._data_to_write)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse("write neighbour memory", "CMD_LINK_WRITE")



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import BinaryIO, Optional
from spinnman.messages.scp.impl import (
    FloodFillEnd, FloodFillStart, FloodFillData)
from spinnman.processes import (
    AbstractMultiConnectionProcess, ConnectionSelector)
from spinnman.constants import UDP_MESSAGE_MAX_SIZE


class WriteMemoryFloodProcess(AbstractMultiConnectionProcess):
    """
    A process for writing memory on multiple SpiNNaker chips at once.
    """
    __slots__ = ()

    def __init__(self, next_connection_selector: ConnectionSelector):
        AbstractMultiConnectionProcess.__init__(
            self, next_connection_selector, n_channels=3,
            intermediate_channel_waits=2)

    def _start_flood_fill(self, n_bytes: int, nearest_neighbour_id: int):
        n_blocks = int(math.ceil(math.ceil(n_bytes / 4.0) /
                                 UDP_MESSAGE_MAX_SIZE))
        with self._collect_responses():
            self._send_request(
                FloodFillStart(nearest_neighbour_id, n_blocks))

    def _end_flood_fill(self, nearest_neighbour_id: int):
        with self._collect_responses():
            self._send_request(FloodFillEnd(nearest_neighbour_id))

    def write_memory_from_bytearray(
            self, nearest_neighbour_id: int, base_address: int,
            data: bytes, offset: int, n_bytes: Optional[int] = None):
        """
        :param int nearest_neighbour_id:
        :param int base_address:
        :param data:
        :type data: bytes or bytearray
        :param int offset:
        :param int n_bytes:
        """
        # pylint: disable=too-many-arguments
        if n_bytes is None:
            n_bytes = len(data)
        self._start_flood_fill(n_bytes, nearest_neighbour_id)

        with self._collect_responses():
            data_offset = offset
            offset = base_address
            block_no = 0
            while n_bytes > 0:
                bytes_to_send = min((int(n_bytes), UDP_MESSAGE_MAX_SIZE))

                self._send_request(FloodFillData(
                    nearest_neighbour_id, block_no, offset,
                    data, data_offset, bytes_to_send))

                block_no += 1
                n_bytes -= bytes_to_send
                offset += bytes_to_send
                data_offset += bytes_to_send

        self._end_flood_fill(nearest_neighbour_id)

    def write_memory_from_reader(
            self, nearest_neighbour_id: int, base_address: int,
            reader: BinaryIO, n_bytes: int):
        """
        :param int nearest_neighbour_id:
        :param int base_address:
        :param reader:
        :type reader: ~io.RawIOBase or ~io.BufferedIOBase
        :param int n_bytes:
        """
        self._start_flood_fill(n_bytes, nearest_neighbour_id)

        with self._collect_responses():
            offset = base_address
            block_no = 0
            while n_bytes > 0:
                bytes_to_send = min((int(n_bytes), UDP_MESSAGE_MAX_SIZE))
                data_array = reader.read(bytes_to_send)

                self._send_request(FloodFillData(
                    nearest_neighbour_id, block_no, offset,
                    data_array, 0, len(data_array)))

                block_no += 1
                n_bytes -= bytes_to_send
                offset += bytes_to_send

        self._end_flood_fill(nearest_neighbour_id)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import cast


class EIEIOType(Enum):
    """
    Possible types of EIEIO packets.
    """
    #: Indicates that data is keys which are 16 bits.
    KEY_16_BIT = (0, 2, 0)
    #: Indicates that data is keys and payloads of 16 bits.
    KEY_PAYLOAD_16_BIT = (1, 2, 2)
    #: Indicates that data is keys of 32 bits.
    KEY_32_BIT = (2, 4, 0)
    #: Indicates that data is keys and payloads of 32 bits.
    KEY_PAYLOAD_32_BIT = (3, 4, 4)

    def __new__(cls, *args) -> 'EIEIOType':
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __init__(self, encoded_value: int,
                 # Default values just to make mypy SHUT UP!
                 # https://github.com/python/mypy/issues/10573
                 key_bytes: int = 0, payload_bytes: int = 0):
        self._encoded_value = encoded_value
        self._key_bytes = key_bytes
        self._payload_bytes = payload_bytes

    @property
    def encoded_value(self) -> int:
        """
        The encoded value representing the type.
        """
        return cast(int, self._value_)

    @property
    def key_bytes(self) -> int:
        """
        The number of bytes used by each key element.

        :rtype: int
        """
        return self._key_bytes

    @property
    def payload_bytes(self) -> int:
        """
        The number of bytes used by each payload element.

        :rtype: int
        """
        return self._payload_bytes

    @property
    def max_value(self) -> int:
        """
        The maximum value of the key or payload (if there is a payload).

        :rtype: int
        """
        return (1 << (self._key_bytes * 8)) - 1



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Iterable, Optional

_MIN_APP_ID = 17
_MAX_APP_ID = 254


class AppIdTracker(object):
    """
    A tracker of application IDs to make it easier to allocate new IDs.
    """
    __slots__ = (
        "_free_ids",
        "_max_app_id",
        "_min_app_id")

    # Keep a class-global reference to the free ID range, so IDs are
    # allocated globally

    def __init__(
            self, app_ids_in_use: Optional[Iterable[int]] = None,
            min_app_id: int = _MIN_APP_ID,
            max_app_id: int = _MAX_APP_ID):
        """
        :param app_ids_in_use: The IDs that are already in use
        :type app_ids_in_use: list(int) or None
        :param int min_app_id: The smallest application ID to use
        :param int max_app_id: The largest application ID to use
        """
        self._free_ids = set(range(min_app_id, max_app_id))
        if app_ids_in_use is not None:
            self._free_ids.difference_update(app_ids_in_use)
        self._min_app_id = min_app_id
        self._max_app_id = max_app_id

    def get_new_id(self) -> int:
        """
        Get a new unallocated ID

        :rtype: int
        """
        return self._free_ids.pop()

    def allocate_id(self, allocated_id: int):
        """
        Allocate a given ID.

        :param int allocated_id: The ID to allocate
        :raises KeyError: If the ID is not present
        """
        self._free_ids.remove(allocated_id)

    def free_id(self, id_to_free: int):
        """
        Free a given ID.

        :param int id_to_free: The ID to free
        :raises KeyError: If the ID is out of range
        """
        if id_to_free < self._min_app_id or id_to_free > self._max_app_id:
            raise KeyError(
                f"ID {id_to_free} out of allowed range of {self._min_app_id} "
                f"to {self._max_app_id}")
        self._free_ids.add(id_to_free)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Optional
from spinnman.data import SpiNNManDataView
from .sdp_flag import SDPFlag

N_BYTES = 8
_EIGHT_BYTES = struct.Struct("<8B")
_SDP_SOURCE_PORT = 7
_SDP_SOURCE_CPU = 31
_SDP_TAG = 0xFF


class SDPHeader(object):
    """
    Represents the header of an SDP message.
    Each optional parameter in the constructor can be set to a value other
    than `None` once, after which it is immutable.  It is an error to set a
    parameter that is not currently `None`.
    """
    __slots__ = (
        "_destination_chip_x",
        "_destination_chip_y",
        "_destination_cpu",
        "_destination_port",
        "_flags",
        "_source_chip_x",
        "_source_chip_y",
        "_source_cpu",
        "_source_port",
        "_tag")

    def __init__(self, flags: SDPFlag, tag: Optional[int] = None,
                 destination_port: int = 0, destination_cpu: int = 0,
                 destination_chip_x: int = 0, destination_chip_y: int = 0,
                 source_port: Optional[int] = None,
                 source_cpu: Optional[int] = None,
                 source_chip_x: Optional[int] = None,
                 source_chip_y: Optional[int] = None):
        """
        :param SDPFlag flags: Any flags for the packet
        :param int tag:
            The IP tag of the packet between 0 and 255, or `None` if it
            is to be set later
        :param int destination_port:
            The destination port of the packet between 0 and 7
        :param int destination_cpu:
            The destination processor ID within the chip between 0 and 31
        :param int destination_chip_x:
            The x-coordinate of the destination chip between 0 and 255
        :param int destination_chip_y:
            The y-coordinate of the destination chip between 0 and 255
        :param int source_port:
            The source port of the packet between 0 and 7, or
            `None` if it is to be set later
        :param int source_cpu:
            The source processor ID within the chip between 0 and 31,
            or `None` if it is to be set later
        :param int source_chip_x:
            The x-coordinate of the source chip between 0 and 255,
            or `None` if it is to be set later
        :param int source_chip_y:
            The y-coordinate of the source chip between 0 and 255,
            or `None` if it is to be set later
        """
        # pylint: disable=too-many-arguments
        self._flags = flags
        self._tag = tag
        self._destination_port = destination_port
        self._destination_cpu = destination_cpu
        self._destination_chip_x = destination_chip_x
        self._destination_chip_y = destination_chip_y
        self._source_port = source_port
        self._source_cpu = source_cpu
        self._source_chip_x = source_chip_x
        self._source_chip_y = source_chip_y

    @property
    def flags(self) -> SDPFlag:
        """
        The flags of the packet (settable).

        :rtype: SDPFlag
        """
        return self._flags

    @flags.setter
    def flags(self, flags: SDPFlag):
        """
        Set the flags of the packet.

        :param SDPFlag flags: The flags to set
        """
        self._flags = flags

    @property
    def tag(self) -> int:
        """
        The tag of the packet, between 0 and 255 (settable).

        :rtype: int
        """
        assert self._tag is not None, "header not yet updated for send"
        return self._tag

    @tag.setter
    def tag(self, tag: int):
        """
        Set the tag of the packet.

        :param int tag: The tag to set, between 0 and 255
        """
        self._tag = tag

    @property
    def destination_port(self) -> int:
        """
        The destination SDP port of the packet, between 0 and 7 (settable).

        :rtype: int
        """
        return self._destination_port

    @destination_port.setter
    def destination_port(self, destination_port: int):
        """
        Set the destination port of the packet.

        :param int destination_port:
            The destination port to set, between 0 and 7
        """
        self._destination_port = destination_port

    @property
    def destination_cpu(self) -> int:
        """
        The core on the destination chip, between 0 and 31 (settable).

        :rtype: int
        """
        return self._destination_cpu

    @destination_cpu.setter
    def destination_cpu(self, destination_cpu: int):
        """
        Set the ID of the destination processor of the packet.

        :param int destination_cpu:
            The processor ID to set, between 0 and 31
        """
        self._destination_cpu = destination_cpu

    @property
    def destination_chip_x(self) -> int:
        """
        The x-coordinate of the destination chip of the packet, between
        0 and 255 (settable).

        :rtype: int
        """
        return self._destination_chip_x

    @destination_chip_x.setter
    def destination_chip_x(self, destination_chip_x: int):
        """
        Set the x-coordinate of the destination chip of the packet.

        :param int destination_chip_x:
            The x-coordinate to set, between 0 and 255
        """
        self._destination_chip_x = destination_chip_x

    @property
    def destination_chip_y(self) -> int:
        """
        The y-coordinate of the destination chip of the packet, between
        0 and 255 (settable).

        :rtype: int
        """
        return self._destination_chip_y

    @destination_chip_y.setter
    def destination_chip_y(self, destination_chip_y: int):
        """
        Set the y-coordinate of the destination chip of the packet.

        :param int destination_chip_y:
            The y-coordinate to set, between 0 and 255
        """
        self._destination_chip_y = destination_chip_y

    @property
    def source_port(self) -> int:
        """
        The source SDP port of the packet, between 0 and 7 (settable).

        :rtype: int
        """
        assert self._source_port is not None, "header not yet updated for send"
        return self._source_port

    @source_port.setter
    def source_port(self, source_port: int):
        """
        Set the source port of the packet.

        :param int source_port: The source port to set, between 0 and 7
        """
        self._source_port = source_port

    @property
    def source_cpu(self) -> int:
        """
        The core on the source chip, between 0 and 31 (settable).

        :rtype: int
        """
        assert self._source_cpu is not None, "header not yet updated for send"
        return self._source_cpu

    @source_cpu.setter
    def source_cpu(self, source_cpu: int):
        """
        Set the ID of the source processor of the packet.

        :param int source_cpu: The processor ID to set, between 0 and 31
        """
        self._source_cpu = source_cpu

    @property
    def source_chip_x(self) -> int:
        """
        The x-coordinate of the source chip of the packet, between
        0 and 255 (settable).

        :rtype: int
        """
        assert self._source_chip_x is not None, \
            "header not yet updated for send"
        return self._source_chip_x

    @source_chip_x.setter
    def source_chip_x(self, source_chip_x: int):
        """
        Set the x-coordinate of the source chip of the packet.

        :param int source_chip_x:
            The x-coordinate to set, between 0 and 255
        """
        self._source_chip_x = source_chip_x

    @property
    def source_chip_y(self) -> int:
        """
        The y-coordinate of the source chip of the packet, between
        0 and 255 (settable).

        :rtype: int
        """
        assert self._source_chip_y is not None, \
            "header not yet updated for send"
        return self._source_chip_y

    @source_chip_y.setter
    def source_chip_y(self, source_chip_y: int):
        """
        Set the y-coordinate of the source chip of the packet.

        :param int source_chip_y:
            The y-coordinate to set, between 0 and 255
        """
        self._source_chip_y = source_chip_y

    @property
    def bytestring(self) -> bytes:
        """
        The header as a byte-string.

        :rtype: bytes
        """
        dest_port_cpu = (((self._destination_port & 0x7) << 5) |
                         (self._destination_cpu & 0x1F))
        source_port_cpu = (((self.source_port & 0x7) << 5) |
                           (self.source_cpu & 0x1F))

        return _EIGHT_BYTES.pack(
            self._flags.value, self.tag, dest_port_cpu, source_port_cpu,
            self._destination_chip_y, self._destination_chip_x,
            self.source_chip_y, self.source_chip_x)

    @staticmethod
    def from_bytestring(data: bytes, offset: int):
        """
        Read the header from a byte-string.

        :param data: The byte-string to read the header from
        :type data: bytes or bytearray
        :param int offset:
            The offset into the data from which to start reading
        """
        (flags, tag, dest_port_cpu, source_port_cpu,
         destination_chip_y, destination_chip_x,
         source_chip_y, source_chip_x) = _EIGHT_BYTES.unpack_from(data, offset)
        destination_port = dest_port_cpu >> 5
        destination_cpu = dest_port_cpu & 0x1F
        source_port = source_port_cpu >> 5
        source_cpu = source_port_cpu & 0x1F
        return SDPHeader(
            SDPFlag(flags), tag, destination_port, destination_cpu,
            destination_chip_x, destination_chip_y, source_port, source_cpu,
            source_chip_x, source_chip_y)

    def get_physical_cpu_id(self) -> str:
        """
        A String describing the physical core of the destination.

        :rtype: str
        """
        return SpiNNManDataView.get_physical_string(
            (self._destination_chip_x, self._destination_chip_y),
            self._destination_cpu)

    def update_for_send(self, source_x: int, source_y: int):
        """
        Apply defaults to the header for sending over UDP.

        :param int source_x:
            Where the packet is deemed to originate: X coordinate
        :param int source_y:
            Where the packet is deemed to originate: Y coordinate
        """
        self.tag = _SDP_TAG
        self.source_port = _SDP_SOURCE_PORT
        self.source_cpu = _SDP_SOURCE_CPU
        self.source_chip_x = source_x
        self.source_chip_y = source_y



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Iterable, Optional
from spinn_utilities.overrides import overrides
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp.enums import SCPCommand
from spinnman.messages.sdp import SDPFlag, SDPHeader
from spinnman.messages.scp import SCPRequestHeader
from .check_ok_response import CheckOKResponse

_WAIT_FLAG = 0x1 << 18


class ApplicationRun(AbstractSCPRequest[CheckOKResponse]):
    """
    An SCP request to run an application loaded on a chip.
    """
    __slots__ = ()

    def __init__(self, app_id: int, x: int, y: int,
                 processors: Optional[Iterable[int]], wait: bool = False):
        """
        :param int app_id: The ID of the application to run, between 16 and 255
        :param int x: The x-coordinate of the chip to run on, between 0 and 255
        :param int y: The y-coordinate of the chip to run on, between 0 and 255
        :param list(int) processors:
            The processors on the chip where the executable should be
            started, between 1 and 17
        :param bool wait:
            True if the processors should enter a "wait" state on starting
        """
        # pylint: disable=too-many-arguments
        processor_mask = 0
        if processors is not None:
            for processor in processors:
                processor_mask |= (1 << processor)

        processor_mask |= (app_id << 24)
        if wait:
            processor_mask |= _WAIT_FLAG

        super().__init__(
            SDPHeader(flags=SDPFlag.REPLY_EXPECTED, destination_port=0,
                      destination_cpu=0, destination_chip_x=x,
                      destination_chip_y=y),
            SCPRequestHeader(command=SCPCommand.CMD_AR),
            argument_1=processor_mask)

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self) -> CheckOKResponse:
        return CheckOKResponse("Run Application", "CMD_AR")



# Copyright (c) 2023 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import (List, Optional)
from spinn_utilities.log import FormatAdapter
from spinn_machine.version.version_3 import Version3
from spinn_machine.version.version_5 import Version5
from spinnman.connections.abstract_classes import Connection
from spinnman.data import SpiNNManDataView
from spinnman.extended.version3transceiver import ExtendedVersion3Transceiver
from spinnman.extended.version5transceiver import ExtendedVersion5Transceiver
from spinnman.model.bmp_connection_data import BMPConnectionData
from spinnman.utilities.utility_functions import (
    work_out_bmp_from_machine_details)
from spinnman.connections.udp_packet_connections import (
    BMPConnection, BootConnection, SCAMPConnection)
from spinnman.transceiver import Transceiver
from spinnman.transceiver.version3transceiver import Version3Transceiver
from spinnman.transceiver.version5transceiver import Version5Transceiver
from spinnman.transceiver.virtual5transceiver import Virtual5Transceiver
from spinnman.constants import LOCAL_HOST

logger = FormatAdapter(logging.getLogger(__name__))


def create_transceiver_from_hostname(
        hostname: Optional[str], *,
        bmp_connection_data: Optional[BMPConnectionData] = None,
        auto_detect_bmp: bool = False, power_cycle: bool = False,
        extended: bool = False) -> Transceiver:
    """
    Create a Transceiver by creating a :py:class:`~.UDPConnection` to the
    given hostname on port 17893 (the default SCAMP port), and a
    :py:class:`~.BootConnection` on port 54321 (the default boot port),
    optionally discovering any additional links using the UDPConnection,
    and then returning the transceiver created with the conjunction of
    the created UDPConnection and the discovered connections.

    :param hostname: The hostname or IP address of the board or `None` if
        only the BMP connections are of interest
    :type hostname: str or None
    :type number_of_boards: int or None
    :param BMPConnectionData bmp_connection_data:
        the details of the BMP connections used to boot multi-board systems
    :param bool auto_detect_bmp:
        ``True`` if the BMP of version 4 or 5 boards should be
        automatically determined from the board IP address
    :param bool power_cycle: If True will power cycle the machine
    :param scamp_connections:
        the list of connections used for SCAMP communications
    :param bool extended:
        If True will return an Extended version of the Transceiver
    :return: The created transceiver
    :rtype: spinnman.transceiver.Transceiver
    :raise SpinnmanIOException:
        If there is an error communicating with the board
    :raise SpinnmanInvalidPacketException:
        If a packet is received that is not in the valid format
    :raise SpinnmanInvalidParameterException:
        If a packet is received that has invalid parameters
    :raise SpinnmanUnexpectedResponseCodeException:
        If a response indicates an error during the exchange
    """
    if hostname is not None:
        logger.info("Creating transceiver for {}", hostname)
    connections: List[Connection] = list()

    # if no BMP has been supplied, but the board is a spinn4 or a spinn5
    # machine, then an assumption can be made that the BMP is at -1 on the
    # final value of the IP address
    version = SpiNNManDataView.get_machine_version()
    if (isinstance(version, Version5) and auto_detect_bmp is True and
            (bmp_connection_data is None)):
        if hostname is None:
            raise ValueError("hostname is required if deriving BMP details")
        bmp_connection_data = \
            work_out_bmp_from_machine_details(hostname)

    # handle BMP connections
    if bmp_connection_data is not None:
        bmp_connection = BMPConnection(bmp_connection_data)
        connections.append(bmp_connection)
        logger.info("Transceiver using BMP: {}",
                    bmp_connection.remote_ip_address)

    connections.append(SCAMPConnection(remote_host=hostname))

    # handle the boot connection
    connections.append(BootConnection(remote_host=hostname))

    if hostname == LOCAL_HOST:
        return create_transceiver_from_connections(
            connections=connections, virtual=True, extended=extended)
    else:
        return create_transceiver_from_connections(
            connections=connections, virtual=False, power_cycle=power_cycle,
            extended=extended)


def create_transceiver_from_connections(
        connections: List[Connection], virtual: bool = False,
        power_cycle: bool = False, extended: bool = False) -> Transceiver:
    """
    Create a Transceiver with these connections

    :param list(Connection) connections:
        An iterable of connections to the board.  If not specified, no
        communication will be possible until connections are found.
    :param bool virtual: If True will return a virtual Transceiver
    :param bool power_cycle: If True will power cycle the machine
    :param bool extended:
    :return: The created transceiver
    :rtype: spinnman.transceiver.Transceiver
    :raise SpinnmanIOException:
        If there is an error communicating with the board
    :raise SpinnmanInvalidPacketException:
        If a packet is received that is not in the valid format
    :raise SpinnmanInvalidParameterException:
        If a packet is received that has invalid parameters
    :raise SpinnmanUnexpectedResponseCodeException:
        If a response indicates an error during the exchange
    """
    version = SpiNNManDataView.get_machine_version()
    if isinstance(version, Version3):
        if virtual:
            raise NotImplementedError(f"No Virtual Transceiver for {version=}")
        if extended:
            return ExtendedVersion3Transceiver(
                connections=connections, power_cycle=power_cycle)
        return Version3Transceiver(
            connections=connections, power_cycle=power_cycle)
    if isinstance(version, Version5):
        if virtual:
            return Virtual5Transceiver(
                connections=connections, power_cycle=power_cycle)
        if extended:
            return ExtendedVersion5Transceiver(
                connections=connections, power_cycle=power_cycle)
        return Version5Transceiver(
            connections=connections, power_cycle=power_cycle)
    raise NotImplementedError(f"No Transceiver for {version=}")



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from spinnman.messages.scp.enums import SCPResult

_TWO_SHORTS = struct.Struct("<2H")


class SCPResponseHeader(object):
    """
    Represents the header of an SCP Response.
    """
    __slots__ = (
        "_result",
        "_sequence")

    def __init__(self, result, sequence):
        """
        :param SCPResult result:
        :param int sequence:
        """
        self._result = result
        self._sequence = sequence

    @property
    def result(self) -> SCPResult:
        """
        The result of the SCP response.

        :rtype: SCPResult
        """
        return self._result

    @property
    def sequence(self) -> int:
        """
        The sequence number of the SCP response, between 0 and 65535.

        :rtype: int
        """
        return self._sequence

    @staticmethod
    def from_bytestring(data: bytes, offset: int) -> 'SCPResponseHeader':
        """
        Read a header from a byte-string.

        :param bytes data: The byte-string to read from
        :param int offset:
        """
        result, sequence = _TWO_SHORTS.unpack_from(data, offset)
        return SCPResponseHeader(SCPResult(result), sequence)



# Copyright (c) 2022 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
API of the client for the Spalloc web service.
"""

import struct
from typing import Optional
from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.overrides import overrides
from spinn_utilities.typing.coords import XY
from spinnman.connections.udp_packet_connections import EIEIOConnection
from spinnman.exceptions import SpinnmanTimeoutException
from spinnman.messages.eieio import (
    AbstractEIEIOMessage,
    read_eieio_command_message, read_eieio_data_message)
from spinnman.messages.sdp import SDPMessage, SDPFlag, SDPHeader
from spinnman.messages.scp.impl import IPTagSet
from .spalloc_proxied_connection import SpallocProxiedConnection

_ONE_SHORT = struct.Struct("<H")
_TWO_SKIP: bytes = b'\0\0'
_NUM_UPDATE_TAG_TRIES = 3
_UPDATE_TAG_TIMEOUT = 1.0


class SpallocEIEIOConnection(
        EIEIOConnection, SpallocProxiedConnection, metaclass=AbstractBase):
    """
    The socket interface supported by proxied EIEIO connected sockets.
    This emulates an :py:class:`EIEOConnection` opened with a remote address
    specified.
    """
    __slots__ = ()

    @overrides(EIEIOConnection.send_eieio_message)
    def send_eieio_message(self, eieio_message: AbstractEIEIOMessage):
        # Not normally used, as packets need headers to go to SpiNNaker
        self.send(eieio_message.bytestring)

    def send_eieio_message_to_core(
            self, eieio_message: AbstractEIEIOMessage, x: int, y: int, p: int):
        """

        :param AbstractEIEIOMessage eieio_message:
        :param int x:
        :param int y:
        :param int p:
        :rtype: None
        """
        sdp_message = SDPMessage(
            SDPHeader(
                flags=SDPFlag.REPLY_NOT_EXPECTED, tag=0,
                destination_port=1, destination_cpu=p,
                destination_chip_x=x, destination_chip_y=y,
                source_port=0, source_cpu=0,
                source_chip_x=0, source_chip_y=0),
            data=eieio_message.bytestring)
        self.send(_TWO_SKIP + sdp_message.bytestring)

    @overrides(EIEIOConnection.receive_eieio_message)
    def receive_eieio_message(
            self, timeout: Optional[float] = None) -> AbstractEIEIOMessage:
        data = self.receive(timeout)
        header = _ONE_SHORT.unpack_from(data)[0]
        if header & 0xC000 == 0x4000:
            return read_eieio_command_message(data, 0)
        return read_eieio_data_message(data, 0)

    @property
    @abstractmethod
    def _coords(self) -> XY:
        """
        The X, Y coordinates of the chip this connection is connected to.

        :rtype: tuple(int,int)
        """
        raise NotImplementedError

    def update_tag(self, tag: int, do_receive: bool = True):
        """
        Update the given tag on the connected Ethernet-enabled chip to send
        messages to this connection.

        :param int tag: The tag ID to update
        :param bool do_receive:
            Whether to do the reception of the response or not
        :raises SpinnmanTimeoutException:
            If the message isn't handled within a reasonable timeout.
        :raises SpinnmanUnexpectedResponseCodeException:
            If the message is rejected by SpiNNaker/SCAMP.
        """
        x, y = self._coords
        request = IPTagSet(
            x, y, [0, 0, 0, 0], 0, tag, strip=True, use_sender=True)
        request.sdp_header.flags = SDPFlag.REPLY_EXPECTED_NO_P2P
        request.sdp_header.update_for_send(x, y)
        data = _TWO_SKIP + request.bytestring
        for _try in range(_NUM_UPDATE_TAG_TRIES):
            try:
                self.send(data)
                if do_receive:
                    response_data = self.receive(_UPDATE_TAG_TIMEOUT)
                    request.get_scp_response().read_bytestring(
                        response_data, len(_TWO_SKIP))
                return
            except SpinnmanTimeoutException as e:
                if _try + 1 == _NUM_UPDATE_TAG_TRIES:
                    raise e



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional, Sequence


class BMPConnectionData(object):
    """
    Contains the details of a BMP connection.
    """
    __slots__ = [
        "_boards",
        "_ip_address",
        "_port_num"]

    def __init__(self, ip_address: str, boards: Sequence[int],
                 port_num: Optional[int]):
        self._ip_address = ip_address
        self._boards = tuple(boards)
        self._port_num = port_num

    @property
    def ip_address(self) -> str:
        """
        The IP address of the BMP.

        :rtype: str
        """
        return self._ip_address

    @property
    def boards(self) -> Sequence[int]:
        """
        The boards to be addressed.

        :rtype: iterable(int)
        """
        return self._boards

    @property
    def port_num(self) -> Optional[int]:
        """
        The port number associated with this BMP connection.

        :return: The port number
        """
        return self._port_num

    def __str__(self) -> str:
        return f"{self._ip_address}:{self._boards}:{self._port_num}"

    def __repr__(self) -> str:
        return self.__str__()



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from spinnman.messages.scp.impl.read_memory import ReadMemory, Response
from spinnman.model import RouterDiagnostics
from .abstract_multi_connection_process import AbstractMultiConnectionProcess
from .abstract_multi_connection_process_connection_selector import (
    ConnectionSelector)

_N_REGISTERS = 16
_ONE_WORD = struct.Struct("<I")


class ReadRouterDiagnosticsProcess(
        AbstractMultiConnectionProcess[Response]):
    """
    A process for reading the diagnostic data block from a SpiNNaker router.
    """
    __slots__ = (
        "_control_register",
        "_error_status",
        "_register_values")

    def __init__(self, connection_selector: ConnectionSelector):
        """
        :param ConnectionSelector connection_selector:
        """
        super().__init__(connection_selector)
        self._control_register = 0
        self._error_status = 0
        self._register_values = [0] * _N_REGISTERS

    def __handle_control_register_response(self, response: Response):
        self._control_register = _ONE_WORD.unpack_from(
            response.data, response.offset)[0]

    def __handle_error_status_response(self, response: Response):
        self._error_status = _ONE_WORD.unpack_from(
            response.data, response.offset)[0]

    def __handle_register_response(self, response: Response):
        for register in range(_N_REGISTERS):
            self._register_values[register] = _ONE_WORD.unpack_from(
                response.data, response.offset + (register * 4))[0]

    def get_router_diagnostics(self, x: int, y: int) -> RouterDiagnostics:
        """
        :param int x:
        :param int y:
        :rtype: RouterDiagnostics
        """
        with self._collect_responses():
            coords = x, y, 0
            self._send_request(ReadMemory(coords, 0xe1000000, 4),
                               self.__handle_control_register_response)
            self._send_request(ReadMemory(coords, 0xe1000014, 4),
                               self.__handle_error_status_response)
            self._send_request(ReadMemory(coords, 0xe1000300, 16 * 4),
                               self.__handle_register_response)

        return RouterDiagnostics(self._control_register, self._error_status,
                                 self._register_values)



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from typing import Iterable, List, Optional
from spinnman.messages.spinnaker_boot import SystemVariableDefinition


class ChipInfo(object):
    """
    Represents the system variables for a chip, received from the chip SDRAM.
    """
    __slots__ = [
        "_ip_address",
        "_led_flash_period_ms",
        "_leds",
        "_links_available",
        "_offset",
        "_physical_to_virtual_core_map",
        "_status_map",
        "_system_data",
        "_virtual_core_ids",
        "_virtual_to_physical_core_map"]

    def __init__(self, system_data: bytes, offset: int):
        """
        :param bytes system_data:
            A byte-string retrieved from SDRAM on the board
        :param int offset:
            The offset into the byte-string where the actual data starts
        :raise SpinnmanInvalidParameterException:
            If the data doesn't contain valid system data information
        """
        self._system_data = system_data
        self._offset = offset

        links_available = self._read_int("links_available")
        self._links_available: List[int] = list()
        for i in range(6):
            if ((links_available >> i) & 0x1) != 0:
                self._links_available.append(i)

        self._led_flash_period_ms = self._read_int(
            "led_half_period_10_ms") * 10
        self._leds = [self._read_int("led_0"), self._read_int("led_1")]
        self._status_map = self._read_bytes("status_map")
        self._physical_to_virtual_core_map = self._read_bytes(
            "physical_to_virtual_core_map")
        self._virtual_to_physical_core_map = self._read_bytes(
            "virtual_to_physical_core_map")

        self._virtual_core_ids: List[int] = list()
        for virtual_core_id in self._physical_to_virtual_core_map:
            if virtual_core_id != 0xFF:
                self._virtual_core_ids.append(virtual_core_id)
        self._virtual_core_ids.sort()

        ip = self._read_bytes("ethernet_ip_address")
        self._ip_address: Optional[str] = f"{ip[0]}.{ip[1]}.{ip[2]}.{ip[3]}"
        if self._ip_address == "0.0.0.0":
            self._ip_address = None

    def _read_int(self, item: str) -> int:
        item_def = SystemVariableDefinition[item]
        code = item_def.data_type.struct_code
        assert item_def.array_size is None
        value, = struct.unpack_from(
            code, self._system_data, self._offset + item_def.offset)
        return value

    def _read_bytes(self, item: str) -> bytes:
        item_def = SystemVariableDefinition[item]
        assert item_def.array_size is not None
        code = f"{item_def.array_size}{item_def.data_type.struct_code}"
        value, = struct.unpack_from(
            code, self._system_data, self._offset + item_def.offset)
        return value

    @property
    def x(self) -> int:
        """
        The X-coordinate of the chip.

        :rtype: int
        """
        return self._read_int("x")

    @property
    def y(self) -> int:
        """
        The Y-coordinate of the chip.

        :rtype: int
        """
        return self._read_int("y")

    @property
    def x_size(self) -> int:
        """
        The number of chips in the X-dimension.

        :rtype: int
        """
        return self._read_int("x_size")

    @property
    def y_size(self) -> int:
        """
        The number of chips in the Y-dimension.

        :rtype: int
        """
        return self._read_int("y_size")

    @property
    def nearest_ethernet_x(self) -> int:
        """
        The X-coordinate of the nearest chip with Ethernet.

        :rtype: int
        """
        return self._read_int("nearest_ethernet_x")

    @property
    def nearest_ethernet_y(self) -> int:
        """
        The Y-coordinate of the nearest chip with Ethernet.

        :rtype: int
        """
        return self._read_int("nearest_ethernet_y")

    @property
    def is_ethernet_available(self) -> bool:
        """
        Whether the Ethernet is running on this chip.

        :rtype: bool
        """
        return self._read_int("is_ethernet_available") == 1

    @property
    def links_available(self) -> Iterable[int]:
        """
        The links that are available on the chip.

        :rtype: iterable(int)
        """
        return self._links_available

    @property
    def cpu_clock_mhz(self) -> int:
        """
        The speed of the CPU clock in MHz.

        :rtype: int
        """
        return self._read_int("cpu_clock_mhz")

    @property
    def physical_to_virtual_core_map(self) -> bytes:
        """
        The physical core ID to virtual core ID map; entries with a value
        of 0xFF are non-operational cores.

        :rtype: bytearray
        """
        return self._physical_to_virtual_core_map

    @property
    def virtual_to_physical_core_map(self) -> bytes:
        """
        The virtual core ID to physical core ID map; entries with a value
        of 0xFF are non-operational cores.

        :rtype: bytearray
        """
        return self._virtual_to_physical_core_map

    @property
    def virtual_core_ids(self) -> Iterable[int]:
        """
        A list of available cores by virtual core ID (including the monitor).

        :rtype: iterable(int)
        """
        return self._virtual_core_ids

    @property
    def sdram_base_address(self) -> int:
        """
        The base address of the user region of SDRAM on the chip.

        :rtype: int
        """
        return self._read_int("sdram_base_address")

    @property
    def system_sdram_base_address(self) -> int:
        """
        The base address of the System SDRAM region on the chip.

        :rtype: int
        """
        return self._read_int("system_sdram_base_address")

    @property
    def cpu_information_base_address(self) -> int:
        """
        The base address of the CPU information structure.

        :rtype: int
        """
        return self._read_int("cpu_information_base_address")

    @property
    def first_free_router_entry(self) -> int:
        """
        The ID of the first free routing entry on the chip.

        :rtype: int
        """
        return self._read_int("first_free_router_entry")

    @property
    def ip_address(self) -> Optional[str]:
        """
        The IP address of the chip, or `None` if no Ethernet.

        :rtype: str
        """
        return self._ip_address

    @property
    def iobuf_size(self) -> int:
        """
        The size of the IOBUF buffers in bytes.

        :rtype: int
        """
        return self._read_int("iobuf_size")

    def router_table_copy_address(self) -> int:
        """
        The address of the copy of the router table.

        :rtype: int
        """
        return self._read_int("router_table_copy_address")

    @property
    def system_ram_heap_address(self) -> int:
        """
        The address of the base of the heap in system RAM.

        :rtype: int
        """
        return self._read_int("system_ram_heap_address")

    @property
    def sdram_heap_address(self) -> int:
        """
        The address of the base of the heap in SDRAM.

        :rtype: int
        """
        return self._read_int("sdram_heap_address")



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class SDPFlag(Enum):
    """
    SDPFlag for the message.
    """
    #: Indicates that a reply is not expected
    REPLY_NOT_EXPECTED = 0x07
    #: Indicates that a reply is expected
    REPLY_EXPECTED = 0x87
    #: Indicates that a reply is not expected and packet should not be
    #: P2P routed.
    REPLY_NOT_EXPECTED_NO_P2P = 0x07 | 0x20
    #: Indicates that a reply is expected and packet should not be P2P routed.
    REPLY_EXPECTED_NO_P2P = 0x87 | 0x20



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional
from .sdp_header import SDPHeader


class SDPMessage(object):
    """
    Wraps up an SDP message with a header and optional data.
    """
    __slots__ = (
        "_data",
        "_offset",
        "_sdp_header")

    def __init__(self, sdp_header: SDPHeader,
                 data: Optional[bytes] = None, offset: int = 0):
        """
        :param SDPHeader sdp_header: The header of the message
        :param data: The data of the SDP packet, or `None` if no data
        :type data: bytes or bytearray or None
        :param int offset: The offset where the valid data starts
        """
        self._sdp_header = sdp_header
        self._data = data
        self._offset = offset

    @property
    def bytestring(self) -> bytes:
        """
        The byte-string of the message.

        :rtype: bytes
        """
        if self._data is not None:
            return self._sdp_header.bytestring + self._data[self._offset:]
        return self._sdp_header.bytestring

    @staticmethod
    def from_bytestring(data: bytes, offset: int) -> 'SDPMessage':
        """
        :param bytes data:
        :param int offset:
        :rtype: SDPMessage
        """
        sdp_header = SDPHeader.from_bytestring(data, offset)
        return SDPMessage(sdp_header, data, offset + 8)

    @property
    def sdp_header(self) -> SDPHeader:
        """
        The header of the packet.

        :rtype: SDPHeader
        """
        return self._sdp_header

    @property
    def data(self) -> Optional[bytes]:
        """
        The data in the packet.

        :rtype: bytes or bytearray or None
        """
        return self._data

    @property
    def offset(self) -> int:
        """
        The offset where the valid data starts.

        :rtype: int
        """
        return self._offset
